# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Query and match objects in Vizier catalogs."""

import copy
import numpy as np
from astroquery.vizier import Vizier, VizierClass

from .base_catalog import _BasePhotometryCatalog
from ._online_tools import astroquery_radius, \
                           astroquery_skycoord


__all__ = ['VizierCatalogClass', 'UCAC5Catalog',
           'UCAC4Catalog', 'GSC23Catalog', 'APASSCalatolg',
           'DENISCatalog', 'TWOMASSCatalog']


class VizierCatalogClass(_BasePhotometryCatalog):
    """Base class to handle with Vizier online catalogs."""

    vizier_table = None
    id_key = None
    ra_key = 'RAJ2000'
    dec_key = 'DEJ2000'
    flux_key = None
    flux_error_key = None
    flux_unit = None
    available_filters = None
    type = 'online'
    prepend_id_key = False
    bibcode = None
    _valid_init_kwargs = set(['vizier_table', 'id_key', 'ra_key', 'dec_key',
                              'flux_key', 'flux_error_key', 'flux_unit',
                              'prepend_id_key', 'available_filters',
                              'bibcode', 'comment'])
    _last_query_info = None
    _last_query_table = None
    _vizier = None

    def __init__(self, **kwargs):
        for i, v in kwargs.items():
            if i in self._valid_init_kwargs:
                self.__setattr__(i, v)
            else:
                raise ValueError('Invalid parameter {} passed to'
                                 ' VizierCatalogClass')

    @property
    def vizier(self):
        """Query operator instance."""
        if self._vizier is None:
            self._vizier = Vizier()
            self._vizier.ROW_LIMIT = -1
        return copy.copy(self._vizier)

    @vizier.setter
    def vizier(self, value):
        if not isinstance(value, VizierClass):
            raise ValueError(f'{value} is not a VizierClass instance.')
        self._vizier = value

    def _flux_keys(self, band):
        flux_key = self.flux_key.format(band=band)
        if self.flux_error_key is not None:
            flux_error_key = self.flux_error_key.format(band=band)
        else:
            flux_error_key = None
        return flux_key, flux_error_key

    def query_object(self, center):
        """Query a single object in the catalog."""
        center = astroquery_skycoord(center)
        return self._query(self.vizier.query_object,
                           center, catalog=self.vizier_table)[0]

    def query_region(self, center, radius):
        """Query all objects in a region."""
        radius = astroquery_radius(radius)
        center = astroquery_skycoord(center)
        viz = self.vizier
        return self._query(viz.query_region,
                           center, radius,
                           catalog=self.vizier_table)[0]

    def _id_resolve(self, idn):
        if self.prepend_id_key:
            if isinstance(self.prepend_id_key, (str, bytes)):
                id_key = self.prepend_id_key
            else:
                id_key = self.id_key
            idn = [f"{id_key} {i}" for i in idn]
            idn = np.array(idn)
        return idn

    def filter_flux(self, band, query=None):
        """Filter the flux data of a query."""
        self.check_filter(band)
        flux_key, flux_error_key = self._flux_keys(band)
        if query is None:
            query = self._last_query_table

        flux = np.array(query[flux_key].data)
        try:
            flux_error = np.array(query[flux_error_key].data)
        except KeyError:
            flux_error = np.array([np.nan]*len(flux))

        return flux, flux_error

    def match_objects(self, ra, dec, band=None, limit_angle='2 arcsec'):
        """Match objects from RA DEC list with this catalog."""
        flux_keys = ['flux', 'flux_error']
        table_props = [('id', ''), ('ra', np.nan), ('dec', np.nan),
                       ('flux', np.nan), ('flux_error', np.nan)]
        res = self._match_objects(ra, dec, band, limit_angle,
                                  flux_keys, table_props)

        return res


UCAC4Catalog = VizierCatalogClass(available_filters=["B", "V", "g", "r", "i"],
                                  vizier_table="I/322A",
                                  prepend_id_key=True,
                                  id_key="UCAC4",
                                  ra_key="RAJ2000",
                                  dec_key="DEJ2000",
                                  flux_key="{band}mag",
                                  flux_error_key="e_{band}mag",
                                  flux_unit="mag",
                                  bibcode="2013AJ....145...44Z",
                                  comment="Magnitudes from APASS")


UCAC5Catalog = VizierCatalogClass(available_filters=["Gaia", "R", "f."],
                                  vizier_table="I/340",
                                  prepend_id_key=True,
                                  id_key='SrcIDgaia',
                                  ra_key="RAJ2000",
                                  dec_key="DEJ2000",
                                  flux_key="{band}mag",
                                  flux_error_key="e_{band}mag",
                                  flux_unit="mag",
                                  bibcode="2017yCat.1340....0Z",
                                  comment="Rmag from NOMAD")


APASSCalatolg = VizierCatalogClass(available_filters=["B", "V", "g'",
                                                      "r'", "i'"],
                                   vizier_table="II/336/apass9",
                                   prepend_id_key=False,
                                   id_key=-1,
                                   ra_key="RAJ2000",
                                   dec_key="DEJ2000",
                                   flux_key="{band}mag",
                                   flux_error_key="e_{band}mag",
                                   flux_unit="mag",
                                   bibcode="2016yCat.2336....0H",
                                   comment="g', r' and i' magnitudes in"
                                           " AB system")


DENISCatalog = VizierCatalogClass(available_filters=["I", "J", "K"],
                                  vizier_table="B/denis",
                                  prepend_id_key=True,
                                  id_key="DENIS",
                                  ra_key="RAJ2000",
                                  dec_key="DEJ2000",
                                  flux_key="{band}mag",
                                  flux_error_key="e_{band}mag",
                                  flux_unit="mag",
                                  bibcode="2005yCat.2263....0T",
                                  comment="Id, Jd, Kd may differ a bit "
                                          "from 2MASS catalog")


TWOMASSCatalog = VizierCatalogClass(available_filters=["I", "J", "K"],
                                    vizier_table="B/denis",
                                    prepend_id_key='2MASS',
                                    id_key="_2MASS",
                                    ra_key="RAJ2000",
                                    dec_key="DEJ2000",
                                    flux_key="{band}mag",
                                    flux_error_key="e_{band}mag",
                                    flux_unit="mag",
                                    bibcode="2003yCat.2246....0C",
                                    comment="")


class _GCS23Catalog(VizierCatalogClass):
    available_filters = ["U", "B", "Bj", "V", "R", "F", "N"]
    vizier_table = "I/305/out"
    prepend_id_key = False
    id_key = 'GSC2.3'
    ra_key = "RAJ2000"
    dec_key = "DEJ2000"
    flux_key = "{band}mag"
    flux_error_key = "e_{band}mag"
    flux_unit = "mag"
    bibcode = "2008AJ....136..735L"
    comment = "Hubble Guide Star Catalog 2.3.2 (STScI, 2006)." \
              " R magnitude was assumed to be equal to F magnitude."
    filt_conversion = {'R': 'F',
                       'Bj': 'j'}

    def _flux_keys(self, band):
        if band in self.filt_conversion.keys():
            band = self.filt_conversion[band]
        return VizierCatalogClass._flux_keys(self, band)


GSC23Catalog = _GCS23Catalog()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import os
import pytest
import numpy as np

from astropy.io import fits
from astropop.file_collection import FitsFileGroup, list_fits_files
from astropop.testing import *


@pytest.fixture(scope='session')
def tmpdir(tmp_path_factory):
    fn = tmp_path_factory.mktemp('filegroups')
    files = {}
    for i in ('fits', 'fz', 'fits.gz', 'fts', 'fit'):
        folder = fn / i.replace('.', '_')
        folder.mkdir()
        files[i] = create_test_files(folder, extension=i)

    # Also create the images on the custom HDU

    tmpdir = fn / 'custom_hdu'
    tmpdir.mkdir()
    flist = []
    for i in range(10):
        fname = tmpdir / f'bias_{i+1}.fits'
        if fname.is_file():
            continue
        hdr = fits.Header({'obstype': 'bias',
                           'exptime': 0.0001,
                           'observer': 'Galileo Galileo',
                           'object': 'bias',
                           'filter': ''})
        hdul = fits.HDUList([
            fits.PrimaryHDU(),
            fits.ImageHDU(np.ones((8, 8), dtype=np.int16), hdr,
                          name='image')
            ])
        hdul.writeto(fname)
        flist.append(str(fname))
    files['custom_hdu'] = flist

    files['custom'] = []
    folder = fn / 'custom'
    folder.mkdir()
    for i in ('myfits', 'otherfits'):
        f = create_test_files(folder, extension=i)
        files['custom'].extend(f)
        files[i] = f

    return fn, files


def create_test_files(tmpdir, extension='fits'):
    """Create dummy test files for testing."""
    files_list = []
    # create 10 bias files
    for i in range(10):
        iname = f'bias_{i}.{extension}'
        fname = tmpdir / iname
        if fname.is_file():
            continue
        hdr = fits.Header({'obstype': 'bias',
                           'exptime': 0.0001,
                           'observer': 'Galileo Galileo',
                           'object': 'bias',
                           'filter': '',
                           'space key': 1,
                           'image': iname})
        hdu = fits.PrimaryHDU(np.ones((8, 8), dtype=np.int16), hdr)
        hdu.writeto(fname)
        files_list.append(str(fname))

    # create 10 flat V files
    for i in range(10):
        iname = f'flat_{i}_v.{extension}'
        fname = tmpdir / iname
        if fname.is_file():
            continue
        hdr = fits.Header({'obstype': 'flat',
                           'exptime': 10.0,
                           'observer': 'Galileo Galileo',
                           'object': 'flat',
                           'filter': 'V',
                           'space key': 1,
                           'image': iname})
        hdu = fits.PrimaryHDU(np.ones((8, 8), dtype=np.int16), hdr)
        hdu.writeto(fname)
        files_list.append(str(fname))

    # create 10 object V files
    for i in range(10):
        iname = f'object_{i}_v.{extension}'
        fname = tmpdir / iname
        if fname.is_file():
            continue
        hdr = fits.Header({'obstype': 'science',
                           'exptime': 1.0,
                           'observer': 'Galileo Galileo',
                           'object': 'Moon',
                           'filter': 'V',
                           'space key': 1,
                           'image': iname})
        hdu = fits.PrimaryHDU(np.ones((8, 8), dtype=np.int16), hdr)
        hdu.writeto(fname)
        files_list.append(str(fname))

    return files_list


class Test_FitsFileGroup():
    def test_fg_creation_empty(self):
        fg = FitsFileGroup()
        assert_is_instance(fg, FitsFileGroup)
        assert_equal(len(fg), 0)

    def test_fg_create_filegroup(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits_gz', compression=True)
        assert_is_instance(fg, FitsFileGroup)
        assert_equal(len(fg), 30)
        assert_equal(sorted(fg.files), sorted(flist['fits.gz']))

    def test_fg_create_filegroup_without_compression(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        assert_is_instance(fg, FitsFileGroup)
        assert_equal(len(fg), 30)
        assert_equal(sorted(fg.files), sorted(flist['fits']))

        #Default is False
        fg = FitsFileGroup(location=tmpdir/'fits')
        assert_is_instance(fg, FitsFileGroup)
        assert_equal(len(fg), 30)
        assert_equal(sorted(fg.files), sorted(flist['fits']))

    def test_fg_creation_files(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(files=flist['fits'])
        assert_is_instance(fg, FitsFileGroup)
        assert_equal(len(fg), 30)
        assert_equal(sorted(fg.files), sorted(flist['fits']))

    def test_fg_creation_no_std_extension(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fts')
        assert_is_instance(fg, FitsFileGroup)
        assert_equal(len(fg), 30)
        assert_equal(sorted(fg.files), sorted(flist['fts']))

        fg = FitsFileGroup(location=tmpdir/'fit')
        assert_is_instance(fg, FitsFileGroup)
        assert_equal(len(fg), 30)
        assert_equal(sorted(fg.files), sorted(flist['fit']))

        fg = FitsFileGroup(location=tmpdir/'fz')
        assert_is_instance(fg, FitsFileGroup)
        assert_equal(len(fg), 30)
        assert_equal(sorted(fg.files), sorted(flist['fz']))

    def test_fg_creation_glob_include_exclude(self, tmpdir):
        tmpdir, flist = tmpdir
        # only bias
        fg = FitsFileGroup(location=tmpdir/'fits', glob_include='*bias*')
        assert_is_instance(fg, FitsFileGroup)
        assert_equal(len(fg), 10)
        assert_equal(sorted(fg.files), sorted(flist['fits'][:10]))

        # everything except bias
        fg = FitsFileGroup(location=tmpdir/'fits', glob_exclude='*bias*')
        assert_is_instance(fg, FitsFileGroup)
        assert_equal(len(fg), 20)
        assert_equal(sorted(fg.files), sorted(flist['fits'][10:]))

    @pytest.mark.parametrize('hdu', [1, 'image'])
    def test_fg_creation_custom_hdu(self, tmpdir, hdu):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=str(tmpdir/'custom_hdu'), ext=hdu)
        assert_is_instance(fg, FitsFileGroup)
        assert_equal(len(fg), 10)
        assert_equal(sorted(fg.files), sorted(flist['custom_hdu']))
        for k in ('object', 'exptime', 'observer', 'filter'):
            assert_in(k, fg.summary.colnames)
        for i in fg.summary:
            assert_equal(i['object'], 'bias')

    def test_fg_filtered_single_key(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        bias_files = flist['fits'][:10]
        flat_files = flist['fits'][10:20]
        sci_files = flist['fits'][20:]

        # object keyword
        fg_b = fg.filtered({'object': 'bias'})
        assert_equal(len(fg_b), 10)
        assert_equal(sorted(fg_b.files),
                     sorted(bias_files))

        # filter keyword
        fg_v = fg.filtered({'filter': 'V'})
        assert_equal(len(fg_v), 20)
        assert_equal(sorted(fg_v.files),
                     sorted(flat_files + sci_files))

        # Hierarch key with space
        fg_s = fg.filtered({'space key': 1})
        assert_equal(len(fg_s), 30)
        assert_equal(sorted(fg_s.files),
                     sorted(bias_files + flat_files + sci_files))

    def test_fg_filtered_multiple_key(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        nfg = fg.filtered({'object': 'Moon',
                           'exptime': 1.0,
                           'image': 'object_4_v.fits'})
        assert_equal(len(nfg), 1)
        assert_equal(nfg.files, [flist['fits'][24]])

    def test_fg_filtered_empty(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)

        # non existing key
        with pytest.raises(KeyError):
            nfg = fg.filtered({'NON-EXISTING': 1})

        # existing but not matched
        nfg = fg.filtered({'object': 'Sun'})
        assert_is_instance(nfg, FitsFileGroup)
        assert_equal(len(nfg), 0)
        assert_equal(nfg.files, [])

    def test_fg_getitem_column(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        obj_column = fg['object']
        assert_equal(sorted(obj_column),
                     sorted(['bias']*10+['flat']*10+['Moon']*10))

    def test_fg_getitem_single_file(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        row = fg[1]
        assert_is_instance(row, FitsFileGroup)
        assert_equal(len(row), 1)
        assert_equal(row.files, [flist['fits'][1]])

    def test_fg_getitem_slice(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        row = fg[2:5]
        assert_is_instance(row, FitsFileGroup)
        assert_equal(len(row), 3)
        assert_equal(row.files, flist['fits'][2:5])

    def test_fg_getitem_array_or_tuple(self, tmpdir):
        tmpdir, flist = tmpdir
        flist = flist['fits']
        files = [flist[2], flist[4]]
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)

        row = fg[[2, 4]]
        assert_is_instance(row, FitsFileGroup)
        assert_equal(len(row), 2)
        assert_equal(row.files, files)

        row = fg[np.array([2, 4])]
        assert_is_instance(row, FitsFileGroup)
        assert_equal(len(row), 2)
        assert_equal(row.files, files)

    def test_fg_getitem_empty(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        row = fg[[]]
        assert_is_instance(row, FitsFileGroup)
        assert_equal(len(row), 0)
        assert_equal(row.files, [])

    def test_fg_getitem_keyerror(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        with pytest.raises(KeyError):
            fg['NonExistingKey']

    def test_fg_setitem_str(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        fg.add_column('new_column')
        fg['new_column'] = ['test']*len(fg)
        assert_equal(sorted(fg['new_column']), ['test']*len(fg))

    def test_fg_setitem_tuple(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        fg.add_column('new_column', values=['test']*len(fg))
        fg['new_column', -1] = 'test1'
        expect = ['test']*len(fg)
        expect[-1] = 'test1'
        assert_equal(sorted(fg['new_column']), expect)

    def test_fg_group_by(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        for i in fg.group_by('object'):
            assert_is_instance(i, FitsFileGroup)
            assert_equal(len(i.values('object', unique=True)), 1)
            assert_in(i.values('object', unique=True)[0],
                      ['bias', 'flat', 'Moon'])
            assert_equal(len(i), 10)

    def test_fg_group_by_used_id(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        fg.add_column('id', values=np.arange(len(fg)))
        for i in fg.group_by('object'):
            assert_is_instance(i, FitsFileGroup)
            assert_equal(len(i.values('object', unique=True)), 1)
            assert_in(i.values('object', unique=True)[0],
                      ['bias', 'flat', 'Moon'])
            assert_equal(len(i), 10)

class Test_ListFitsFiles():
    def test_list_custom_extension(self, tmpdir):
        tmpdir, flist = tmpdir
        found_files = list_fits_files(tmpdir/'custom',
                                      fits_extensions='.myfits')
        assert_equal(sorted(found_files), sorted(flist['myfits']))

        found_files = list_fits_files(tmpdir/'custom',
                                      fits_extensions=['.myfits', '.otherfits'])
        assert_equal(sorted(found_files), sorted(flist['custom']))

    @pytest.mark.parametrize('ext', ['fits', 'fz', 'fit', 'fts'])
    def test_list_no_extension(self, tmpdir, ext):
        tmpdir, flist = tmpdir
        found_files = list_fits_files(tmpdir/f'{ext}')
        assert_equal(sorted(found_files), sorted(flist[ext]))

    def test_list_glob_include(self, tmpdir):
        tmpdir, flist = tmpdir
        found_files = list_fits_files(tmpdir/'fits', glob_include='*bias*')
        # must find only bias
        assert_equal(sorted(found_files), sorted(flist['fits'][:10]))

    def test_list_glob_exclude(self, tmpdir):
        tmpdir, flist = tmpdir
        found_files = list_fits_files(tmpdir/'fits', glob_exclude='*bias*')
        # find everything except bias
        assert_equal(sorted(found_files), sorted(flist['fits'][10:]))


class Test_FitsFileGroup_Paths():
    def test_fg_full_path_file(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        for i in range(len(fg)):
            assert_equal(fg.full_path(i), str(flist['fits'][i]))
        for f in fg.files:
            assert_equal(fg.full_path(f), str(f))

    def test_fg_full_path_file_db(self, tmpdir):
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False,
                           database=tmpdir/'fits'/'database.db')

        for i in range(len(fg)):
            assert_equal(fg.full_path(i), str(flist['fits'][i]))
        for f in fg.files:
            assert_equal(fg.full_path(f), str(f))

        # but the files must be stored relativelly in database
        for i in range(len(fg)):
            assert_equal(fg['__file'][i],
                         os.path.relpath(flist['fits'][i], tmpdir/'fits'))

    def test_fg_relative_path_file(self, tmpdir):
        # no database means full paths
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False)
        for i in range(len(fg)):
            assert_equal(fg.relative_path(flist['fits'][i]), flist['fits'][i])

        for f in fg.files:
            assert_equal(fg.relative_path(f), f)

    def test_fg_relative_path_file_db(self, tmpdir):
        # database means relative paths
        tmpdir, flist = tmpdir
        fg = FitsFileGroup(location=tmpdir/'fits', compression=False,
                           database=tmpdir/'fits'/'database.db')
        for i in range(len(fg)):
            assert_equal(fg.relative_path(flist['fits'][i]),
                         os.path.relpath(flist['fits'][i], tmpdir/'fits'))

        for f in fg.files:
            assert_equal(fg.relative_path(f),
                         os.path.relpath(f, tmpdir/'fits'))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
'''Functions to easy handle with astronomical coordinates.'''

import numpy as np
from astropy.coordinates import SkyCoord
from astropy import units
from ..py_utils import check_iterable

__all__ = ['guess_coordinates']


def guess_coordinates(ra, dec, skycoord=True):
    """Try to guess the format or ra and dec passed."""
    # TODO: this needs a refactor!
    # process as lists of coordinates
    if check_iterable(ra) and check_iterable(dec):
        if len(ra) != len(dec):
            raise ValueError('RA and Dec do not match in dimensions.')
        try:
            ra = np.array(ra, dtype='f8')
            dec = np.array(dec, dtype='f8')
            s = None
        except ValueError:
            s = SkyCoord(ra, dec, unit=(units.hourangle, units.deg))
    elif check_iterable(ra) or check_iterable(dec):
        raise ValueError('RA and Dec do not match in dimensions.')

    # process as single coordinates
    else:
        try:
            ra = float(ra)
            dec = float(dec)
            s = None
        except ValueError:
            # Assume (at least for now) that it's in sexagesimal
            s = SkyCoord(ra, dec, unit=(units.hourangle, units.deg))

    if s is not None:
        if skycoord:
            return s
        return s.ra.degree, s.dec.degree
    else:
        if skycoord:
            return SkyCoord(ra, dec, unit=(units.deg, units.deg))
        return ra, dec



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
from astropop.math.physical import units
from astropop._unit_property import unit_property
from astropop.testing import *


@unit_property
class DummyClass():
    def __init__(self, unit):
        self.unit = unit


@pytest.mark.parametrize('unit,expect', [('meter', units.m),
                                         (units.adu, units.adu),
                                         (None, units.dimensionless_unscaled),
                                         ('', units.dimensionless_unscaled)])
def test_qfloat_unit_property(unit, expect):
    # Getter test
    c = DummyClass(unit)
    assert_equal(c.unit, expect)

    # Setter test
    c = DummyClass(None)
    c.unit = unit
    assert_equal(c.unit, expect)


def test_qfloat_unit_property_none():
    # Check None and dimensionless_unscaled
    c = DummyClass(None)
    assert_is_none(c._unit)
    assert_equal(c.unit, units.dimensionless_unscaled)


@pytest.mark.parametrize('unit', ['None', 'Invalid'])
def test_qfloat_unit_property_invalid(unit):
    with pytest.raises(ValueError):
        DummyClass(unit)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Math operations with uncertainties and units.

Simplified version of `uncertainties` python package with some
`~astropy.units` addings, in a much more free form.
"""

import numbers
from functools import partial
from astropy import units
from astropy.units.quantity_helper.helpers import get_converters_and_unit
from astropy.units import UnitsError, Quantity
import numpy as np

from ..py_utils import check_iterable
from .._unit_property import unit_property
from ._deriv import propagate_2, propagate_1


__all__ = ['unit_property', 'QFloat', 'qfloat', 'units', 'UnitsError',
           'equal_within_errors', 'UnitsError']


HANDLED_AFUNCS = {}
HANDLED_UFUNCS = {}  # must be func(method, *inputs, **kwargs)


def _implements_array_func(numpy_function):
    """Register an __array_function__ implementation for QFloat objects."""
    def decorator_array_func(func):
        HANDLED_AFUNCS[numpy_function] = func
        return func
    return decorator_array_func


def _implements_ufunc(numpy_ufunc):
    """Register an ufunc implementation for QFloat objects."""
    def decorator_ufunc(func):
        HANDLED_UFUNCS[numpy_ufunc] = func
        return func
    return decorator_ufunc


def convert_to_qfloat(value):
    """Convert a value to QFloat.

    Notes
    -----
    - The `unit` is extracted from a `unit` attribute in the number.
      If this attribute is not present, the number is considered dimensionless.
    - Compilant classes now are:
        * Python standard numbers;
        * Numpy simple arrays.
    """
    # Not change if value is already a qfloat
    if isinstance(value, QFloat):
        return value

    # extract unit (force)
    unit = getattr(value, 'unit', None)

    # number support
    if isinstance(value, numbers.Number):
        # Pure numbers or NDArray don't have uncertainty
        return QFloat(value, None, unit)

    # Numpy arrays support. They need to handle single numbers or ufloat
    if isinstance(value, (np.ndarray, list, tuple)):
        # Astropy Quantities
        if isinstance(value, Quantity):
            return QFloat(value.value, None, value.unit)
        # Everithing else is considered numbers.
        return QFloat(value, None, unit)

    # FrameData support
    if value.__class__.__name__ == 'FrameData':
        # avoid breaking the code due to cyclic imports
        return QFloat(value.data, value.get_uncertainty(False), value.unit)

    # Handle Astropy units to multuply
    if isinstance(value, (units.UnitBase, str)):
        return QFloat(1.0, 0.0, value)

    # TODO: astropy NDData support?

    raise ValueError(f'Value {value} is not QFloat compilant.')


def require_qfloat(func):
    """Require qfloat as argument decorator."""
    def decorator(self, *others):
        others = [convert_to_qfloat(i) for i in others]
        return func(self, *others)
    return decorator


class _QFloatFormatter():
    """Simple store for numbers that can be rounded to first digit.

    Used mainly for speedup repr and str.
    """
    _rounded = False

    def __init__(self, nominal, std):
        self._n = nominal
        self._s = std
        self._d = np.nan

    @property
    def nominal(self):
        if not self._rounded:
            self.round()
        return self._n

    @property
    def std_dev(self):
        if not self._rounded:
            self.round()
        return self._s

    @property
    def digits(self):
        if not self._rounded:
            self.round()
        return self._d

    def round(self):
        try:
            first_digit = -np.int(np.floor(np.log10(np.abs(self._s))))
            self._n = np.around(self._n, first_digit)
            self._s = np.around(self._s, first_digit)
            self._d = first_digit
        except (ValueError, ZeroDivisionError, OverflowError):
            # Do not change the values
            pass

    def __format__(self, format_spec):
        # For not, format do not matter
        # Positive digits, decimals. Negative digits, integers.
        nominal = self.nominal
        std = self.std_dev
        digits = self.digits

        if not np.isnan(digits):
            if digits > 0:
                n_part = f"{nominal:.{int(digits)}f}"
                s_part = f"{std:.{int(digits)}f}"
            else:
                n_part = f"{nominal:.0f}"
                s_part = f"{std:.0f}"
            return f"{n_part}+-{s_part}"

        return f"{nominal}+-{std}"

    def __repr__(self):
        return f"{self}"


def _create_formater(nominal, std):
    """Create _QFloatFormater handling lists."""
    if check_iterable(nominal):
        return [_create_formater(n, s) for n, s in zip(nominal, std)]
    return _QFloatFormatter(nominal, std)


def same_unit(qfloat1, qfloat2, func=None):
    """Put 2 qfloats in the same unit."""
    # both units must be the same
    def convert(converter, qf, unit):
        if converter is None:
            return qf
        nom = converter(qf.nominal)
        std = converter(qf.uncertainty)
        return QFloat(nom, std, unit)

    qfloat1, qfloat2 = [convert_to_qfloat(i) for i in (qfloat1, qfloat2)]

    # The error raising require a funcion name
    converters, unit = get_converters_and_unit(func, qfloat1.unit,
                                               qfloat2.unit)
    qfloat1 = convert(converters[0], qfloat1, unit)
    qfloat2 = convert(converters[1], qfloat2, unit)

    return qfloat1, qfloat2


def equal_within_errors(qf1, qf2):
    """Check if two QFloats are equal within errors.

    Parameters
    ----------
    qf1, qf2: `~astropop.math.QFloat`, `float` or `np.ndarray`
        QFloats to compare.

    Returns
    -------
    bool:
        `True` if the numbers are equal within the uncertainties,
        (the difference is smaller then the sum of errors). `False`
        if they are different.

    Notes
    -----
    - We consider two numbers equal within errors when
      number1 - number2 <= error1 + error2
    - Incompatible units means different numbers.
    """
    qf1, qf2 = [convert_to_qfloat(i) for i in (qf1, qf2)]
    try:
        qf1, qf2 = same_unit(qf1, qf2, equal_within_errors)
    except UnitsError:
        # Incompatible units are different numbers.
        return False

    diff = np.abs(qf1.nominal - qf2.nominal)
    erro = qf1.uncertainty + qf2.uncertainty

    return diff <= erro


def qfloat(value, uncertainty=None, unit=None):
    """Create a QFloat from the values.

    Parameters
    ----------
    value: number or array_like
        Nominal value(s) of the quantity.
    uncertainty : number, array_like or `None` (optional)
        Uncertainty value of the quantity. If `None`, the quantity will be
        considered with no errors. Must match `value` shape.
    unit: `~astropy.units.Unit` or string (optional)
        The data unit. Must be `~astropy.units.Unit` compliant.

    Returns
    -------
    f: `~astropop.math.physical.QFloat`
        The QFloat created.
    """
    f = QFloat(value, uncertainty, unit)
    return f


@unit_property
class QFloat():
    """Storing float values with stddev uncertainties and units.

    Parameters
    ----------
    value : number, `~astropop.math.QFloat` or array_like
        Nominal value(s) of the quantity. Must be a real number, array of real
        numbers or a QFloat.
    uncertainty : number, array_like or `None` (optional)
        Uncertainty value of the quantity. If `None`, the quantity will be
        considered with no errors. Must match `value` shape.
    unit : `~astropy.units.Unit` or string (optional)
        The data unit. Must be `~astropy.units.Unit` compliant.

    Notes
    -----
    - This class don't support memmapping. Is intended to be in memory ops.
    - Units are handled by `~astropy.units`.
    - Math operations cares about units and uncertainties.
    """

    _nominal = None
    _uncert = None
    _unit = None

    def __init__(self, value, uncertainty=None, unit=None):
        value, uncertainty, unit = self._check_inputs(value, uncertainty, unit)
        self._nominal = value
        self._set_uncert(uncertainty)
        self.unit = unit

    def _check_inputs(self, value, uncertainty=None, unit=None):
        if isinstance(value, QFloat):
            qf = value
            value = value.nominal
            if uncertainty is not None:
                raise ValueError('uncertainty must be None if value is a '
                                 'QFloat.')
            uncertainty = qf.uncertainty
            if unit is not None:
                raise ValueError('unit must be None if value is a QFloat.')
            unit = qf.unit
        if np.any(np.array(value) == None):  # noqa: E711
            raise TypeError('value must be not None.')
        for i in value, uncertainty:
            if not np.any(np.isreal(i)):
                raise TypeError('value and uncertainty must be real numbers, '
                                'or arrays of real numbers.')
        value = np.array(value) if check_iterable(value) else float(value)
        return value, uncertainty, unit

    def _set_uncert(self, value):
        if value is None:
            if check_iterable(self._nominal):
                self._uncert = np.zeros_like(self._nominal)
            else:
                self._uncert = 0.0
        else:
            if not np.any(np.isreal(value)):
                raise TypeError('uncertainty must be real numbers')
            if np.shape(value) != np.shape(self._nominal):
                raise ValueError('Uncertainty with shape different from '
                                 'nominal value: '
                                 f'{np.shape(value)} '
                                 f'{np.shape(self._nominal)}')
            if check_iterable(self._nominal):
                # Errors must be always positive
                value = np.array(value)
                value[value == None] = 0.0  # noqa: E711
                self._uncert = np.abs(np.array(value))
            else:
                self._uncert = float(abs(value))

    def _set_nominal(self, value):
        uncertainty = None
        unit = None
        if isinstance(value, tuple):
            if len(value) == 2:
                value, uncertainty = value
            else:
                value, uncertainty, unit = value
        value, uncertainty, unit = self._check_inputs(value, uncertainty, unit)
        self._nominal = value
        self._set_uncert(uncertainty)
        if unit is not None:
            self.unit = unit

    @property
    def uncertainty(self):
        """Uncertainty of the quantity."""
        return self._uncert

    @uncertainty.setter
    def uncertainty(self, value):
        self._set_uncert(value)

    @property
    def nominal(self):
        """Nominal value of the quantity."""
        return self._nominal

    @nominal.setter
    def nominal(self, value):
        self._set_nominal(value)

    @property
    def std_dev(self):
        """Alias for uncertainty."""
        return self.uncertainty

    @std_dev.setter
    def std_dev(self, value):
        self.uncertainty = value

    @property
    def shape(self):
        return np.shape(self.nominal)

    @property
    def size(self):
        return np.size(self.nominal)

    def reset(self, value, uncertainty=None, unit=None):
        """Reset all the data.

        Parameters
        ----------
        value : number or array_like
            Nominal value(s) of the quantity.
        uncertainty : number, array_like or `None` (optional)
            Uncertainty value of the quantity. If `None`, the quantity will be
            considered with no errors. Must match `value` shape.
        unit : `~astropy.units.Unit` or string (optional)
            The data unit. Must be `~astropy.units.Unit` compliant.
        """
        self.nominal = value
        self.uncertainty = uncertainty
        self.unit = unit

    def to(self, unit):
        """Convert this QFloat to another unit.

        Parameters
        ----------
        - unit: string or `~astropy.units.UnitBase`
            Unit to converto to.

        Returns
        -------
        - QFloat:
            A new instance of this class, converted to the new unit.
        """
        other = units.Unit(unit, parse_strict='silent')
        (_, conv), unit = get_converters_and_unit(self.to, other, self.unit)
        if conv is not None:
            nvalue = conv(self.nominal)
            nstd = conv(self.uncertainty)
            return QFloat(nvalue, nstd, unit)
        # None converter means no conversion
        return QFloat(self.nominal, self.uncertainty, self.unit)

    @property
    def value(self):
        """Same as nominal. For Astropy compatibility."""
        return self.nominal

    def __repr__(self):
        # FIXME: repr can be very slow for mutch large arrays
        # repr for arrays
        if check_iterable(self.nominal):
            ret = "<QFloat\n"
            ret2 = _create_formater(self.nominal, self.std_dev)
            ret2 = np.array(ret2).__repr__()
            ret2 += f'\nunit={str(self.unit)}'
        # repr for single values
        else:
            ret = "<QFloat "
            ret2 = f"{_QFloatFormatter(self.nominal, self.std_dev)}"
            ret2 += f' {str(self.unit)}'
        ret += ret2 + '>'
        return ret

    def __getitem__(self, index):
        """Get one item of given index IF this is iterable."""
        v = self.nominal[index]
        s = self.uncertainty[index]
        return QFloat(v, s, self.unit)

    def __setitem__(self, index, value):
        """Set one item at given index if this is iterable."""
        value = convert_to_qfloat(value)
        _, value = same_unit(self, value, self.__setitem__)

        self._nominal[index] = value.nominal
        self._uncert[index] = value.uncertainty

    def __len__(self):
        return len(self.nominal)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Wrap numpy ufuncs, using uncertainties and units.

        Parameters
        ----------
        function : callable
            Ufunc object that was called.
        method : str
            String indicating which Ufunc method was called
            (``__call__``, ``reduce``, ``reduceat``, ``accumulate``,
            ``outer`` or ``inner``).
        inputs : tuple
            A tuple of the input arguments to the ``ufunc``.
        kwargs : keyword arguments
            A dictionary containing the optional input arguments of the
            ``ufunc``. If given, any out arguments, both positional and
            keyword, are passed as a ``tuple`` in kwargs.

        Returns
        -------
        result : `~astropop.math.QFloat`
            Results of the ufunc, with the unit and uncertainty.
        """
        # Only call supported now
        if method != '__call__':
            return NotImplemented

        if ufunc not in HANDLED_UFUNCS:
            return NotImplemented

        # put all inputs as QFloats, a local "require_qfloat"
        inputs = [convert_to_qfloat(i) for i in inputs]

        out = kwargs.get('out', None)
        if out is not None:
            raise NotImplementedError("`out` argument not supported yet.")

        result = HANDLED_UFUNCS[ufunc](*inputs, **kwargs)

        return result

    def __array_function__(self, func, types, args, kwargs):
        """Wrap numpy functions.

        Parameters
        ----------
        func: callable
            Arbitrary callable exposed by NumPy’s public API.
        types: list
            Collection of unique argument types from the original NumPy
            function call that implement ``__array_function__``.
        args: tuple
            Positional arguments directly passed on from the original call.
        kwargs: dict
            Keyword arguments directly passed on from the original call.
        """
        if func not in HANDLED_AFUNCS:
            return NotImplemented

        return HANDLED_AFUNCS[func](*args, **kwargs)

    def __eq__(self, other):
        try:
            this, other = same_unit(self, other, self.__eq__)
        except Exception:
            # Incompatible units are different numbers.
            # Incompatible types are different
            return False

        if np.any(this.nominal != other.nominal) or \
           np.any(this.uncertainty != other.uncertainty):
            return False
        return True

    def __ne__(self, other):
        return not self == other

    @require_qfloat
    def __gt__(self, other):
        this, other = same_unit(self, other, self.__gt__)
        return this.nominal > other.nominal

    @require_qfloat
    def __ge__(self, other):
        this, other = same_unit(self, other, self.__lt__)
        return this.nominal >= other.nominal

    @require_qfloat
    def __lt__(self, other):
        this, other = same_unit(self, other, self.__lt__)
        return this.nominal < other.nominal

    @require_qfloat
    def __le__(self, other):
        this, other = same_unit(self, other, self.__lt__)
        return this.nominal <= other.nominal

    def __lshift__(self, other):
        """Lshift operator used to convert units."""
        return self.to(other)

    def __ilshift__(self, other):
        res = self << other
        self.reset(res.nominal, res.uncertainty, res.unit)
        return self

    @require_qfloat
    def __add__(self, other):
        qf1, qf2 = same_unit(self, other, self.__add__)
        sum_n = qf1.nominal + qf2.nominal
        sum_s = propagate_2('add', sum_n,
                            qf1.nominal, qf2.nominal,
                            qf1.std_dev, qf2.std_dev)
        return QFloat(sum_n, sum_s, qf1.unit)

    @require_qfloat
    def __iadd__(self, other):
        res = self.__add__(other)
        self.reset(res.nominal, res.uncertainty, res.unit)
        return self

    @require_qfloat
    def __radd__(self, other):
        return other.__add__(self)

    @require_qfloat
    def __sub__(self, other):
        qf1, qf2 = same_unit(self, other, self.__add__)
        sub_n = qf1.nominal - qf2.nominal
        sub_s = propagate_2('add', sub_n,
                            qf1.nominal, qf2.nominal,
                            qf1.std_dev, qf2.std_dev)
        return QFloat(sub_n, sub_s, qf1.unit)

    @require_qfloat
    def __isub__(self, other):
        res = self.__sub__(other)
        self.reset(res.nominal, res.uncertainty, res.unit)
        return self

    @require_qfloat
    def __rsub__(self, other):
        return -self.__sub__(other)

    @require_qfloat
    def __mul__(self, other):
        unit = self.unit * other.unit
        qf1, qf2 = self, other
        mul_n = qf1.nominal * qf2.nominal
        mul_s = propagate_2('mul', mul_n,
                            qf1.nominal, qf2.nominal,
                            qf1.std_dev, qf2.std_dev)
        return QFloat(mul_n, mul_s, unit)

    @require_qfloat
    def __imul__(self, other):
        res = self.__mul__(other)
        self.reset(res.nominal, res.uncertainty, res.unit)
        return self

    @require_qfloat
    def __rmul__(self, other):
        return self.__mul__(other)

    @require_qfloat
    def __truediv__(self, other):
        unit = self.unit / other.unit
        qf1, qf2 = self, other
        div_n = qf1.nominal / qf2.nominal
        div_s = propagate_2('truediv', div_n,
                            qf1.nominal, qf2.nominal,
                            qf1.std_dev, qf2.std_dev)
        return QFloat(div_n, div_s, unit)

    @require_qfloat
    def __itruediv__(self, other):
        res = self.__truediv__(other)
        self.reset(res.nominal, res.uncertainty, res.unit)
        return self

    @require_qfloat
    def __rtruediv__(self, other):
        return other.__truediv__(self)

    @require_qfloat
    def __floordiv__(self, other):
        unit = self.unit / other.unit
        qf1, qf2 = self, other
        div_n = qf1.nominal // qf2.nominal
        div_s = propagate_2('floordiv', div_n,
                            qf1.nominal, qf2.nominal,
                            qf1.std_dev, qf2.std_dev)
        return QFloat(div_n, div_s, unit)

    @require_qfloat
    def __ifloordiv__(self, other):
        res = self.__floordiv__(other)
        self.reset(res.nominal, res.uncertainty, res.unit)
        return self

    @require_qfloat
    def __rfloordiv__(self, other):
        # As the argument always enter here as a qfloat...
        return other.__floordiv__(self)

    @require_qfloat
    def __mod__(self, other):
        qf1, qf2 = self, other
        mod_n = qf1.nominal % qf2.nominal
        mod_s = propagate_2('mod', mod_n,
                            qf1.nominal, qf2.nominal,
                            qf1.std_dev, qf2.std_dev)
        return QFloat(mod_n, mod_s, self.unit)

    @require_qfloat
    def __imod__(self, other):
        res = self.__mod__(other)
        self.reset(res.nominal, res.uncertainty, res.unit)
        return self

    @require_qfloat
    def __rmod__(self, other):
        return other.__mod__(self)

    @require_qfloat
    def __pow__(self, other):
        # Important: different from normal arrays, QFloat cannot be raises
        # to an array due to inconsistencies in unit. Each element could
        # have it's own unit.
        if other.unit != units.dimensionless_unscaled or \
           not np.isscalar(other.nominal):
            raise ValueError('Power operation size-1 require dimensionless'
                             ' expoent')
        qf1, qf2 = self, other
        pow_n = qf1.nominal ** qf2.nominal
        pow_s = propagate_2('pow', pow_n,
                            qf1.nominal, qf2.nominal,
                            qf1.std_dev, qf2.std_dev)
        return QFloat(pow_n, pow_s, self.unit**other.nominal)

    @require_qfloat
    def __ipow__(self, other):
        res = self.__pow__(other)
        self.reset(res.nominal, res.uncertainty, res.unit)
        return self

    @require_qfloat
    def __rpow__(self, other):
        return other.__pow__(self)

    @require_qfloat
    def __neg__(self):
        return QFloat(-self.nominal, self.uncertainty, self.unit)

    @require_qfloat
    def __pos__(self):
        return QFloat(self.nominal, self.uncertainty, self.unit)

    @require_qfloat
    def __abs__(self):
        return QFloat(np.abs(self.nominal), self.uncertainty, self.unit)

    @require_qfloat
    def __int__(self):
        return np.int(self.nominal)

    @require_qfloat
    def __float__(self):
        return np.float(self.nominal)


# TODO:
# Array functions:
#             - copyto, broadcast, broadcast_to
#             - sum, prod, nanprod, nansum, cumprod, cumsum, nancumprod,
#             - nancumsum, diff, ediff1d, cross
#             - concatenate, stack, block, vstack, hstack, dstack, columnstack

@_implements_array_func(np.shape)
def _qfloat_shape(qf):
    """Implement np.shape for qfloats."""
    return qf.shape


@_implements_array_func(np.size)
def _qfloat_size(qf):
    return qf.size


@_implements_array_func(np.clip)
def _qfloat_clip(qf, a_min, a_max, **kwargs):
    # we mantain the original errors
    nominal = np.clip(qf.nominal, a_min, a_max)
    return QFloat(nominal, qf.std_dev, qf.unit)


# Use a simple wrapper for general functions
def _array_func_simple_wrapper(numpy_func):
    """Wraps simple array functions.

    Notes
    -----
    - Functions elegible for these are that ones who applies for nominal and
      std_dev values and return a new QFloat with the applied values.
    - No conversion or special treatment is done in this wrapper.
    - Only for one array ate once.
    """
    def wrapper(qf, *args, **kwargs):
        nominal = numpy_func(qf.nominal, *args, **kwargs)
        std = numpy_func(qf.uncertainty, *args, **kwargs)
        return QFloat(nominal, std, qf.unit)
    _implements_array_func(numpy_func)(wrapper)


_array_func_simple_wrapper(np.delete)
_array_func_simple_wrapper(np.expand_dims)
_array_func_simple_wrapper(np.flip)
_array_func_simple_wrapper(np.fliplr)
_array_func_simple_wrapper(np.flipud)
_array_func_simple_wrapper(np.moveaxis)
_array_func_simple_wrapper(np.ravel)
_array_func_simple_wrapper(np.repeat)
_array_func_simple_wrapper(np.reshape)
_array_func_simple_wrapper(np.resize)
_array_func_simple_wrapper(np.roll)
_array_func_simple_wrapper(np.rollaxis)
_array_func_simple_wrapper(np.rot90)
_array_func_simple_wrapper(np.squeeze)
_array_func_simple_wrapper(np.swapaxes)
_array_func_simple_wrapper(np.take)
_array_func_simple_wrapper(np.tile)
_array_func_simple_wrapper(np.transpose)


@_implements_array_func(np.round)
@_implements_array_func(np.around)
def _qfloat_round(qf, decimals=0):
    """Implement np.round for qfloats."""
    nominal = np.round(qf.nominal, decimals)
    std = np.round(qf.uncertainty, decimals)
    return QFloat(nominal, std, qf.unit)


@_implements_array_func(np.append)
def _qfloat_append(qf, values, axis=None):
    """Implement np.append for qfloats."""
    # First, convert to the same unit.
    qf1, qf2 = same_unit(qf, values, func=np.append)
    nominal = np.append(qf1.nominal, qf2.nominal, axis)
    std = np.append(qf1.uncertainty, qf2.uncertainty, axis)
    return QFloat(nominal, std, qf1.unit)


@_implements_array_func(np.insert)
def _qfloat_insert(qf, obj, values, axis=None):
    """Implement np.insert for qfloats."""
    # Same unit needed too
    qf1, qf2 = same_unit(qf, values, func=np.insert)
    nominal = np.insert(qf1.nominal, obj, qf2.nominal, axis)
    std = np.insert(qf1.uncertainty, obj, qf2.uncertainty, axis)
    return QFloat(nominal, std, qf1.unit)


@_implements_array_func(np.sum)
def _qfloat_sum(qf, axis=None):
    """Implement np.sum for qfloats."""
    nominal = np.sum(qf.nominal, axis=axis)
    std = np.sqrt(np.sum(np.square(qf.uncertainty), axis=axis))
    return QFloat(nominal, std, qf.unit)


@_implements_array_func(np.mean)
def _qfloat_mean(qf, axis=None):
    """Implement np.mean for qfloats."""
    nominal = np.mean(qf.nominal, axis=axis)
    # error of average = std_dev/sqrt(N)
    std = np.std(qf.nominal, axis=axis)
    # N is determined by the number of elements in the axis
    std /= np.sqrt(np.sum(np.ones_like(qf.nominal), axis=axis))
    return QFloat(nominal, std, qf.unit)


def _implements_ufunc_on_nominal(func):
    """Wraps ufuncs only on the nominal value and don't return QFloat."""
    def wrapper(qf, *args, **kwargs):
        return func(qf.nominal, *args, **kwargs)
    _implements_ufunc(func)(wrapper)


_implements_ufunc_on_nominal(np.isnan)
_implements_ufunc_on_nominal(np.isinf)
_implements_ufunc_on_nominal(np.isfinite)
_implements_ufunc_on_nominal(np.isneginf)
_implements_ufunc_on_nominal(np.isposinf)
_implements_ufunc_on_nominal(np.isreal)
_implements_ufunc_on_nominal(np.iscomplex)
_implements_ufunc_on_nominal(np.isscalar)
_implements_ufunc_on_nominal(np.signbit)
_implements_ufunc_on_nominal(np.sign)


def _qfloat_exp_log(qf, func):
    """General implementation for exp and log functions."""
    if qf.unit != units.dimensionless_unscaled:
        raise UnitsError(f'{func.__name__} is only defined for dimensionless'
                         ' quantities.')
    val = func(qf.nominal)
    std = propagate_1(func.__name__, val, qf.nominal, qf.uncertainty)
    return QFloat(val, std, units.dimensionless_unscaled)


_implements_ufunc(np.exp)(partial(_qfloat_exp_log, func=np.exp))
_implements_ufunc(np.exp2)(partial(_qfloat_exp_log, func=np.exp2))
_implements_ufunc(np.expm1)(partial(_qfloat_exp_log, func=np.expm1))
_implements_ufunc(np.log)(partial(_qfloat_exp_log, func=np.log))
_implements_ufunc(np.log2)(partial(_qfloat_exp_log, func=np.log2))
_implements_ufunc(np.log10)(partial(_qfloat_exp_log, func=np.log10))
_implements_ufunc(np.log1p)(partial(_qfloat_exp_log, func=np.log1p))


def _qfloat_floor_wrapper(qf, func):
    """General implementation for floor, ceil and trunc."""
    return QFloat(func(qf.nominal), np.round(qf.uncertainty, 0), qf.unit)


_implements_ufunc(np.floor)(partial(_qfloat_floor_wrapper, func=np.floor))
_implements_ufunc(np.ceil)(partial(_qfloat_floor_wrapper, func=np.ceil))
_implements_ufunc(np.trunc)(partial(_qfloat_floor_wrapper, func=np.trunc))


def _qfloat_only_nominal_wrapper(qf, func):
    """General implementation for isfinite, isinf, isnan."""
    return func(qf.nominal)


_implements_ufunc(np.isfinite)(partial(_qfloat_only_nominal_wrapper,
                                       func=np.isfinite))
_implements_ufunc(np.isinf)(partial(_qfloat_only_nominal_wrapper,
                                    func=np.isinf))
_implements_ufunc(np.isnan)(partial(_qfloat_only_nominal_wrapper,
                                    func=np.isnan))


@_implements_ufunc(np.radians)
@_implements_ufunc(np.deg2rad)
def _qfloat_radians(qf, *args, **kwargs):
    """Convert any qfloat angle to radian."""
    return qf.to(units.radian)


@_implements_ufunc(np.degrees)
@_implements_ufunc(np.rad2deg)
def _qfloat_degrees(qf, *args, **kwargs):
    return qf.to(units.degree)


def _trigonometric_simple_wrapper(numpy_ufunc):
    def trig_wrapper(qf, *args, **kwargs):
        # check if qf is angle
        if qf.unit not in (units.degree, units.radian):
            raise UnitsError('qfloat unit is not degree or radian.')

        # if degree, convert to radian as required for numpy inputs.
        if qf.unit == units.degree:
            qf = qf.to(units.radian)

        nominal = numpy_ufunc(qf.nominal)
        std = propagate_1(numpy_ufunc.__name__, nominal,
                          qf.nominal, qf.std_dev)
        return QFloat(nominal, std, units.dimensionless_unscaled)
    _implements_ufunc(numpy_ufunc)(trig_wrapper)


def _inverse_trigonometric_simple_wrapper(numpy_ufunc):
    def inv_wrapper(qf, *args, **kwargs):
        if qf.unit != units.dimensionless_unscaled:
            raise UnitsError('inverse trigonometric functions require '
                             'dimensionless unscaled variables.')

        nominal = numpy_ufunc(qf.nominal)
        std = propagate_1(numpy_ufunc.__name__, nominal,
                          qf.nominal, qf.std_dev)

        return QFloat(nominal, std, units.radian)
    _implements_ufunc(numpy_ufunc)(inv_wrapper)


_trigonometric_simple_wrapper(np.sin)
_trigonometric_simple_wrapper(np.cos)
_trigonometric_simple_wrapper(np.tan)
_trigonometric_simple_wrapper(np.sinh)
_trigonometric_simple_wrapper(np.cosh)
_trigonometric_simple_wrapper(np.tanh)
_inverse_trigonometric_simple_wrapper(np.arcsin)
_inverse_trigonometric_simple_wrapper(np.arccos)
_inverse_trigonometric_simple_wrapper(np.arctan)
_inverse_trigonometric_simple_wrapper(np.arcsinh)
_inverse_trigonometric_simple_wrapper(np.arccosh)
_inverse_trigonometric_simple_wrapper(np.arctanh)


@_implements_ufunc(np.arctan2)
def _qfloat_arctan2(qf1, qf2):
    """Compute the arctangent of qf1/qf2."""
    # The 2 values must be in the same unit.
    qf2 = qf2.to(qf1.unit)
    nominal = np.arctan2(qf1.nominal, qf2.nominal)
    std = propagate_2('arctan2', nominal, qf1.nominal, qf2.nominal,
                      qf1.std_dev, qf2.std_dev)
    return QFloat(nominal, std, units.radian)


_ufunc_translate = {
    'add': QFloat.__add__,
    'absolute': QFloat.__abs__,
    'divide': QFloat.__truediv__,
    'float_power': QFloat.__pow__,
    'floor_divide': QFloat.__floordiv__,
    'multiply': QFloat.__mul__,
    'negative': QFloat.__neg__,
    'positive': QFloat.__pos__,
    'power': QFloat.__pow__,
    'mod': QFloat.__mod__,
    'remainder': QFloat.__mod__,
    'subtract': QFloat.__sub__,
    'true_divide': QFloat.__truediv__,
    'divmod': lambda x, y: (QFloat.__floordiv__(x, y), QFloat.__mod__(x, y)),
}


def _general_ufunc_wrapper(numpy_ufunc):
    """Implement ufuncs for general math operations.

    Notes
    -----
    - These functions will not operate with kwarg.
    - These functions will just wrap QFloat math methods.
    """
    ufunc_name = numpy_ufunc.__name__
    true_func = _ufunc_translate[ufunc_name]

    def ufunc_wrapper(*inputs):
        return true_func(*inputs)
    _implements_ufunc(numpy_ufunc)(ufunc_wrapper)


_general_ufunc_wrapper(np.add)
_general_ufunc_wrapper(np.absolute)
_general_ufunc_wrapper(np.divide)
_general_ufunc_wrapper(np.divmod)
_general_ufunc_wrapper(np.float_power)
_general_ufunc_wrapper(np.floor_divide)
_general_ufunc_wrapper(np.mod)
_general_ufunc_wrapper(np.multiply)
_general_ufunc_wrapper(np.negative)
_general_ufunc_wrapper(np.positive)
_general_ufunc_wrapper(np.power)
_general_ufunc_wrapper(np.remainder)
_general_ufunc_wrapper(np.subtract)
_general_ufunc_wrapper(np.true_divide)


@_implements_ufunc(np.square)
def _qfloat_square(qf):
    return qf * qf


@_implements_ufunc(np.sqrt)
def _qfloat_sqrt(qf):
    return qf**0.5


@_implements_ufunc(np.hypot)
def _qfloat_hypot(qf1, qf2):
    qf2 = qf2.to(qf1.unit)
    nominal = np.hypot(qf1.nominal, qf2.nominal)
    std = propagate_2('hypot', nominal, qf1.nominal, qf2.nominal,
                      qf1.std_dev, qf2.std_dev)
    return QFloat(nominal, std, qf1.unit)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Dynamic memmap arrays, that can be enabled or disabled."""

import os
import numpy as np

from .compat import EmptyDataError


__all__ = ['MemMapArray', 'create_array_memmap', 'delete_array_memmap']


redirects = ['flags', 'shape', 'strides', 'ndim', 'data', 'size',
             'itemsize', 'nbytes', 'base', 'dtype', 'T', 'real',
             'imag', 'flat', 'ctypes', 'item', 'tolist', 'itemset',
             'tostring', 'tobytes', 'tofile', 'dump', 'dumps', 'astype',
             'byteswap', 'copy', 'view', 'getfield', 'setflags',
             'reshape', 'resize', 'transpose', 'swapaxes', 'flatten',
             'ravel', 'squeeze', 'take', 'put', 'repeat', 'choose', 'sort',
             'argsort', 'partition', 'argpartition', 'searchsorted', 'nonzero',
             'compress', 'diagonal', 'max', 'argmax', 'min', 'argmin', 'ptp',
             'conj', 'round', 'trace', 'sum', 'cumsum', 'mean', 'var', 'std',
             'prod', 'cumprod', 'all', 'any']


def create_array_memmap(filename, data, dtype=None):
    """Create a memory map to an array data.

    Parameters
    ----------
    filename : `string`, `~pathlib.Path`
        Name of memmap file to be created.
    data : array_like
        Data to be stored in the memmap.
    dtype : `string`, `~numpy.dtype` or `None` (optional)
        `~numpy.dtype` compilant data type. If `None`, `data.dtype` will be
        used.

    Returns
    -------
    memmap : `~numpy.memmap`
        Memmap object of cached data.
    """
    if data is None:
        return None

    if filename is None:
        raise ValueError('Could not create a memmap file with None filename.')

    dtype = dtype or data.dtype
    shape = data.shape
    if data.ndim > 0:
        memmap = np.memmap(filename, mode='w+', dtype=dtype, shape=shape)
        memmap[:] = data[:]
    else:
        memmap = data
    return memmap


def delete_array_memmap(memmap, read=True, remove=False):
    """Delete a memmap and read the data to a np.ndarray.

    Parameters
    ----------
    memmap : array_like or `None`
        MemMap array to be deleted.
    read : `bool` (optional)
        Read the data to memory before delete.
    remove : `bool` (optional)
        Delete the memmap file.

    Returns
    -------
    data : array_like or `None`
        Data read from memmap.
    """
    if memmap is None or np.shape(memmap) == ():
        return None

    if read:
        data = np.array(memmap[:])
    else:
        data = None

    if not isinstance(memmap, np.memmap):
        return data

    name = memmap.filename
    if remove:
        del memmap
        if os.path.exists(name):
            os.remove(name)
    return data


def to_memmap_operator(item):
    """Wrap operators to `MemMapArray`."""
    def wrapper(self, *args, **kwargs):
        if not self.empty:
            func = self._contained.__getattribute__(item)
            func = to_memmap_attr(func)
            return func(*args, **kwargs)
        raise EmptyDataError('Empty data container.')
    return wrapper


def to_memmap_attr(func):
    """Wrap attrs to `MemMapArray`."""
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if isinstance(result, np.ndarray):
            return MemMapArray(result, memmap=False)
        return result
    return wrapper


class MemMapArray:
    """An array object that can be easily cached (memmaped) to disk.

    `MemMapArray` is a wrapper around `~numpy.nddata` that allows easy enable
    or disable of memmapping feature (using `~numpy.memmap`). It also wraps
    all numpy array operations and must behave like every single
    `numpy.ndarray`.

    Parameters
    ----------
    data : array_like or `None`
        The data to be assigned to array. If `None`, an empty array will be
        created.
    filename : `string`, `~pathlib.Path` or `None` (optional)
        Base file name for the cache file.
    dtype : `string`, `~numpy.dtype` or `None` (optional)
        Data type to be used in the array to be stores. If `None`, altomatic
        `numpy` dtype will be used. Must be `numpy.dtype` compilant.
    memmap : `bool` (optional)
        If the instance is set to memmap mode from the start.
        Default: `False`

    Notes
    -----
    - This is just a numeric storing array. No unit is assigned and it behaves
      like any dimensionless number in operations.
    """
    # TODO: __copy__

    _filename = None  # filename of memmap
    _file_lock = False  # lock filename
    _contained = None  # Data contained: numpy ndarray or memmap
    _memmap = False

    def __init__(self, data, filename=None, dtype=None,
                 memmap=False):
        # None data should generate a empty container
        if data is None:
            self._contained = None
        else:
            if not hasattr(data, 'dtype'):
                data = np.array(data)
            dtype = dtype or data.dtype
            # Default dtype
            dtype = dtype or 'float64'
            self._contained = np.array(data, dtype=dtype)
            # ensure native byteorder
            if not self._contained.dtype.isnative:
                self._contained = self._contained.byteswap().newbyteorder()
        self.set_filename(filename)
        self._file_lock = True

        if memmap:
            self.enable_memmap()

    @property  # read only
    def empty(self):
        """True if contained data is empty (None)."""
        return self._contained is None

    @property  # read only
    def filename(self):
        """Name of the file where data is cached, if memmap enabled."""
        if self.memmap and self._contained is not None:
            return self._contained.filename
        return self._filename

    @property  # read unly
    def memmap(self):
        """True if memmap is enabled."""
        return self._memmap

    def set_filename(self, value):
        """Set the memmap filename.

        If the instance is already memmapping, memmap will be moved to the new
        file.

        Parameters
        ----------
        value : `string`
            New memmap filename.
        """
        if not self._file_lock:
            self._filename = value
        elif value != self._filename:
            raise ValueError('Filename locked.')

        if self.memmap:
            if value != self._contained.filename:
                n_mm = create_array_memmap(value, self._contained)
                delete_array_memmap(self._contained, read=False, remove=True)
                self._contained = n_mm

    def enable_memmap(self, filename=None):
        """Enable data file memmapping (write data to disk).

        Parameters
        ----------
        filename : `string` or `None` (optional)
            File name of new memmapping. If `None`, the class default
            value will be used.
        """
        if self.memmap:
            return

        if filename is not None:
            self.set_filename(filename)

        self._contained = create_array_memmap(self._filename, self._contained)
        self._memmap = True

    def disable_memmap(self, remove=False):
        """Disable data file memmapping (read to memory).

        Parameters
        ----------
        remove : `bool`
            Remove the memmap file after read values.
        """
        if not self.memmap:
            return

        self._contained = delete_array_memmap(self._contained, read=True,
                                              remove=remove)
        self._memmap = False

    def flush(self):
        """Write changes to disk if memmapping."""
        if self.memmap:
            self._contained.flush()

    def reset_data(self, data=None, dtype=None):
        """Set new data.

        Parameters
        ----------
        data : `~numpy.ndarray` or `None` (optional)
            Data array to be owned. If None, the container will be empty.
        dtype : `string` or `numpy.dtype` (optional)
            Imposed data type.
        """
        if data is None:
            if self.memmap:
                # Need to delete memmap
                mm = self._contained
                delete_array_memmap(mm, read=False, remove=False)
                self._contained = None
            else:
                # don't need to delete memmap
                self._contained = None

        # Not None data
        else:
            adata = data
            if isinstance(data, MemMapArray):
                adata = data._contained
            if not hasattr(adata, 'dtype'):
                adata = np.array(adata)
            if not adata.dtype.isnative:
                adata = adata.byteswap().newbyteorder()
                dtype = adata.dtype
            if self.memmap:
                name = self.filename
                mm = self._contained
                self._contained = create_array_memmap(name, adata, dtype)
                delete_array_memmap(mm, read=False, remove=True)
            else:
                self._contained = np.array(adata, dtype=dtype)

    def __getitem__(self, item):
        if self.empty:
            raise EmptyDataError('Empty data contaier')

        # This cannot create a new MemMapArray to don't break a[x][y] = z
        result = self._contained[item]
        return result

    def __setitem__(self, item, value):
        if self.empty:
            raise EmptyDataError('Empty data container')
        self._contained[item] = value

    def __getattribute__(self, item):
        if item in redirects and not self.empty:
            attr = getattr(self._contained, item)
            if callable(attr):
                attr = to_memmap_attr(attr)
            elif isinstance(attr, np.ndarray):
                attr = MemMapArray(attr, memmap=False)
            return attr
        if item in redirects and self.empty:
            raise EmptyDataError('Empty data container')

        return object.__getattribute__(self, item)

    def __repr__(self):
        return 'MemMapArray:\n' + repr(self._contained) + \
               f'\nfile: {self.filename}'

    def __array__(self, dtype=None):
        if self.empty:
            return np.array(None)
        # Ignore memmapping
        return np.array(self._contained, dtype=dtype)

    def __del__(self):
        """Safe destruct the MemMapArray."""
        try:
            os.remove(self._filename)
        except (FileNotFoundError, TypeError):
            pass

    __lt__ = to_memmap_operator('__lt__')
    __le__ = to_memmap_operator('__le__')
    __gt__ = to_memmap_operator('__gt__')
    __ge__ = to_memmap_operator('__ge__')
    __eq__ = to_memmap_operator('__eq__')
    __ne__ = to_memmap_operator('__ne__')
    __add__ = to_memmap_operator('__add__')
    __sub__ = to_memmap_operator('__sub__')
    __mul__ = to_memmap_operator('__mul__')
    __pow__ = to_memmap_operator('__pow__')
    __truediv__ = to_memmap_operator('__truediv__')
    __floordiv__ = to_memmap_operator('__floordiv__')
    __mod__ = to_memmap_operator('__mod__')
    __lshift__ = to_memmap_operator('__lshift__')
    __rshift__ = to_memmap_operator('__rshift__')
    __and__ = to_memmap_operator('__and__')
    __or__ = to_memmap_operator('__or__')
    __xor__ = to_memmap_operator('__xor__')
    __neg__ = to_memmap_operator('__neg__')
    __pos__ = to_memmap_operator('__pos__')
    __abs__ = to_memmap_operator('__abs__')
    __invert__ = to_memmap_operator('__invert__')
    __matmul__ = to_memmap_operator('__matmul__')
    __bool__ = to_memmap_operator('__bool__')
    __float__ = to_memmap_operator('__float__')
    __complex__ = to_memmap_operator('__complex__')
    __int__ = to_memmap_operator('__int__')
    __iadd__ = to_memmap_operator('__iadd__')
    __isub__ = to_memmap_operator('__isub__')
    __ipow__ = to_memmap_operator('__ipow__')
    __imul__ = to_memmap_operator('__imul__')
    __itruediv__ = to_memmap_operator('__itruediv__')
    __ifloordiv__ = to_memmap_operator('__ifloordiv__')
    __imod__ = to_memmap_operator('__imod__')
    __ilshift__ = to_memmap_operator('__ilshift__')
    __irshift__ = to_memmap_operator('__irshift__')
    __iand__ = to_memmap_operator('__iand__')
    __ior__ = to_memmap_operator('__ior__')
    __ixor__ = to_memmap_operator('__ixor__')
    __len__ = to_memmap_operator('__len__')
    __contains__ = to_memmap_operator('__contains__')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest

from astropop.pipelines import Manager, Config, Instrument, Stage, Product
from astropop.pipelines import FrozenError
from astropop.testing import *

string_store = []


class DummyInstrument(Instrument):
    a = 'a+b='
    b = 'b*d='
    _identifier = 'test.instrument'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def sum_numbers(self, a, b):
        return a+b

    def multiply_numbers(self, b, d):
        return b*d

    def gen_string(self, ab, bd):
        return f"{self.a}{ab} {self.b}{bd}"


class SumStage(Stage):
    _default_config = dict(a=2, b=3, c=1)
    _required_variables = ['d']
    _requested_functions = ['sum_numbers', 'multiply_numbers']
    _provided = ['dummy_sum', 'a_c_number', 'dummy_mult']
    _raise_error = True

    @staticmethod
    def callback(instrument, variables, config, logger):
        a = config.get('a')
        b = config.get('b')
        c = config.get('c')
        d = variables.get('d')

        logger.warning('Testing Warnings')

        s = instrument.sum_numbers(a, b)
        m = instrument.multiply_numbers(b, d)

        return {'dummy_sum': s,
                'a_c_number': c,
                'dummy_mult': m}


class StringStage(Stage):
    _default_config = dict(c_str='c=')
    _required_variables = ['dummy_sum', 'a_c_number']
    _optional_variables = ['dummy_mult']
    _requested_functions = ['gen_string']
    _provided = ['string_c', 'string_abbd']
    _raise_error = True

    @staticmethod
    def callback(instrument, variables, config, logger):
        logger.debug(config)
        c_str = config.get('c_str')
        c = variables.get('a_c_number')
        s = variables.get('dummy_sum')
        m = variables.get('dummy_mult')

        string_c = f"{c_str}{c}"
        string_abbd = instrument.gen_string(s, m)

        return {'string_c': string_c,
                'string_abbd': string_abbd}


class GlobalStage(Stage):
    _default_config = dict()
    _required_variables = ['string_c', 'string_abbd']
    _raise_error = True

    @staticmethod
    def callback(instrument, variables, config, logger):
        cs = variables.get('string_c')
        ad = variables.get('string_abbd')

        global string_store
        string_store.append((cs, ad))

        return {}


class TestManager(Manager):
    config = Config(stages={'sum': {'a': 1, 'b': 2, 'c': 3}})

    def setup_pipeline(self):
        self.register_stage('sum', SumStage(self.factory))
        self.register_stage('string', StringStage(self.factory))
        self.register_stage('globalvars', GlobalStage(self.factory))

    def setup_products(self, name, d):
        i = DummyInstrument()
        p = Product(manager=self, instrument=i)
        self.add_product(name, p)
        p.add_target('globalvars')
        self.set_value(p, 'd', d)


def test_pipeline_complete_flow():
    global string_store
    string_store = []
    m = TestManager()
    m.setup_pipeline()
    m.setup_products('first_product', 4)
    m.setup_products('secon_product', 8)
    m.show_products()
    m.run()
    m.run(index=0, target='sum')
    assert_equal(string_store, [('c=3', 'a+b=3 b*d=16'),
                                ('c=3', 'a+b=3 b*d=16')])


def test_config_item_freezing():
    with pytest.raises(FrozenError):
        c = Config()
        c.freeze()
        c['testing_value'] = 1


def test_config_item_del_freezing():
    with pytest.raises(FrozenError):
        c = Config()
        c['test'] = 1
        c.freeze()
        del c['test']


def test_config_attr_freezing():
    with pytest.raises(FrozenError):
        c = Config()
        c.freeze()
        c.test = 1


def test_config_attr_del_freezing():
    with pytest.raises(FrozenError):
        c = Config()
        c.test = 1
        c.freeze()
        del c.test


def test_instrument_attr_freezing():
    with pytest.raises(FrozenError):
        c = DummyInstrument()
        c.freeze()
        c.test = 1


def test_instrument_attr_del_freezing():
    with pytest.raises(FrozenError):
        c = DummyInstrument()
        c.test = 1
        c.freeze()
        del c.test



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import tempfile
from astropop.testing import (assert_equal, assert_not_equal,
                              assert_almost_equal,
                              assert_true, assert_false,
                              assert_is, assert_is_not,
                              assert_in, assert_not_in,
                              assert_is_none, assert_is_not_none,
                              assert_greater, assert_greater_equal,
                              assert_less, assert_less_equal,
                              assert_is_instance, assert_is_not_instance,
                              assert_path_exists, assert_path_not_exists)
import numpy as np
import pytest


class TestTestingHelpers():
    """Test the testing helpers."""

    def test_numpy_err_msg(self):
        # Ensure compatibility with numpy testing
        with pytest.raises(AssertionError):
            assert_almost_equal([1, 2], [3, 4], 0, None)
        with pytest.raises(AssertionError):
            assert_equal([1, 2], [3, 4], None)
        with pytest.raises(AssertionError):
            assert_almost_equal([1, 2], [3, 4], 0, '')
        with pytest.raises(AssertionError):
            assert_equal([1, 2], [3, 4], '')

    def test_assert_equal(self):
        assert_equal(1, 1)
        assert_equal(np.arange(4), [0, 1, 2, 3])
        assert_equal([np.nan, np.inf, 0, 1], [np.nan, np.inf, 0, 1])

        with pytest.raises(AssertionError):
            assert_equal(1, 2)
        with pytest.raises(AssertionError):
            assert_equal(np.arange(5), [0, 1, 2, 3])
        with pytest.raises(AssertionError):
            assert_equal([np.nan, 0, 1], [0, 0, 1])

    def test_assert_not_equal(self):
        assert_not_equal(1, 2)
        assert_not_equal(np.arange(5), [0, 1, 2, 3])
        assert_not_equal([np.nan, 0, 1], [0, 0, 1])

        with pytest.raises(AssertionError):
            assert_not_equal(1, 1)
        with pytest.raises(AssertionError):
            assert_not_equal(np.arange(4), [0, 1, 2, 3])
        with pytest.raises(AssertionError):
            assert_not_equal([np.nan, np.inf, 0, 1], [np.nan, np.inf, 0, 1])

    def test_assert_almost_equal(self):
        assert_almost_equal(1.0578602, 1.0578603)
        with pytest.raises(AssertionError):
            assert_almost_equal(1, 2)

    def test_assert_true(self):
        assert_true(True)
        with pytest.raises(AssertionError):
            assert_true(False)

    def test_assert_false(self):
        assert_false(False)
        with pytest.raises(AssertionError):
            assert_false(True)

    def test_assert_is(self):
        a = [1]
        b = a
        assert_is(a, b)
        with pytest.raises(AssertionError):
            assert_is(a, [])

    def test_assert_is_not(self):
        a = [1]
        b = [1]
        assert_is_not(a, b)
        with pytest.raises(AssertionError):
            c = a
            assert_is_not(a, c)

    def test_assert_is_none(self):
        a = None
        assert_is_none(a)
        with pytest.raises(AssertionError):
            assert_is_none(1)

    def test_assert_is_not_none(self):
        a = 0
        assert_is_not_none(a)
        with pytest.raises(AssertionError):
            b = None
            assert_is_not_none(b)

    def test_assert_is_instance(self):
        a = []
        assert_is_instance(a, list)
        with pytest.raises(AssertionError):
            assert_is_instance(a, dict)

    def test_assert_is_not_instance(self):
        a = []
        assert_is_not_instance(a, dict)
        with pytest.raises(AssertionError):
            assert_is_not_instance(a, list)

    def test_assert_in(self):
        assert_in(1, [1, 2, 3, 4, 5])
        assert_in('a', 'abc')
        with pytest.raises(AssertionError):
            assert_in('d', 'abc')
        with pytest.raises(AssertionError):
            assert_in(1, [2, 3, 4])

    def test_assert_not_in(self):
        assert_not_in(6, [1, 2, 3, 4, 5])
        assert_not_in('d', 'abc')
        with pytest.raises(AssertionError):
            assert_not_in('a', 'abc')
        with pytest.raises(AssertionError):
            assert_not_in(2, [2, 3, 4])

    def test_assert_greater(self):
        assert_greater(3, 2)
        with pytest.raises(AssertionError):
            assert_greater(3, 3)
            assert_greater(2, 3)

    def test_assert_greater_equal(self):
        assert_greater_equal(3, 2)
        assert_greater_equal(3, 3)
        with pytest.raises(AssertionError):
            assert_greater_equal(2, 3)

    def test_assert_less(self):
        assert_less(1, 2)
        with pytest.raises(AssertionError):
            assert_less(3, 3)
        with pytest.raises(AssertionError):
            assert_less(4, 3)

    def test_assert_less_equal(self):
        assert_less_equal(2, 3)
        assert_less_equal(3, 3)
        with pytest.raises(AssertionError):
            assert_less_equal(4, 3)

    def test_assert_path_exists(self):
        f = tempfile.mkstemp()[1]
        assert_path_exists(f)

    def test_assert_path_not_exists(self):
        f = 'this_path_do_not_exists_c1234123@3412341233456efsccdwwefxsd'
        assert_path_not_exists(f)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Module to handle spherical coordinates reprojection."""

from astropy.wcs import WCS
from astropy.coordinates import SkyCoord
import numpy as np


def gnomonic_tangential_projection(coords, center, pixel_scale=1./3600.):
    """Project a list of coordinates using gnomonic tangential projetion.

    Parameters
    ----------
    coords : `~astropy.coordinates.SkyCoord` or list
        List of coordinates to project. If a list is passed, it must be (n, 2)
        dimention containing decimal degrees RA and DEC coordinates.
    center : `~astropy.coordinates.SkyCoord` or list
        Coordinates of the center ofthe tangential plane. If a list, it must
        have length 2 and contain decimal degrees of RA and DEC coordinates.
    pixel_scale : float (optional)
        Angular size of each pixel (or unit) in the tangential plane.
        Default: 1./3600. (1 arcsec/px)

    Returns
    -------
    list :
        list containing (x, y) coordinates projected in the plane for each
        input position. (n, 2) dimension.
    """
    if isinstance(coords, (list, tuple, np.ndarray)):
        coords = SkyCoord(coords, unit='deg')
    if not isinstance(coords, SkyCoord):
        raise TypeError('coords must by SkyCoord compatible.')

    if not isinstance(center, SkyCoord):
        center = SkyCoord(*center, unit='deg')

    center = [center.ra.degree, center.dec.degree]

    w = WCS(naxis=2)
    w.wcs.crpix = [0, 0]
    w.wcs.crval = center
    w.wcs.ctype = ['RA---TAN', 'DEC--TAN']
    w.wcs.cdelt = [pixel_scale, pixel_scale]

    return np.transpose(coords.to_pixel(w))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Compute polarimetry of dual beam polarimeters images."""

import abc
import numpy as np
from dataclasses import dataclass
from astropy.table import Table
from scipy.spatial import cKDTree
from scipy.optimize import curve_fit
from astropy import units
from functools import partial

from ..logger import logger
from ..math.physical import QFloat


__all__ = ['estimate_dxdy', 'match_pairs',
           'quarterwave_model', 'halfwave_model']


def _compute_theta(q, u):
    """Compute theta using Q and U, considering quadrants and max 180 value."""
    # numpy arctan2 already looks for quadrants and is defined in [-pi, pi]
    theta = np.degrees(0.5*np.arctan2(u, q))
    if not hasattr(theta, 'unit'):
        theta = theta*units.degree
    # do not allow negative values
    if theta < 0*units.degree:
        theta += 180*units.degree
    return theta


def estimate_dxdy(x, y, steps=[100, 30, 5, 3], bins=30, dist_limit=100):
    """Estimate the displacement between the two beams.

    To compute the displacement between the ordinary and extraordinary
    beams, this function computes the most common distances between the
    sources in image, using clipped histograms around the peak.
    """
    def _find_max(d):
        dx = 0
        for lim in (np.max(d), *steps):
            lo, hi = (dx-lim, dx+lim)
            lo, hi = (lo, hi) if (lo < hi) else (hi, lo)
            histx = np.histogram(d, bins=bins, range=[lo, hi])
            mx = np.argmax(histx[0])
            dx = (histx[1][mx]+histx[1][mx+1])/2
        return dx

    # take all combinations
    comb = np.array(np.meshgrid(np.arange(len(x)),
                                np.arange(len(x)))).T.reshape(-1, 2)
    # filter only y[j] > y[i]
    filt = y[comb[:, 1]] > y[comb[:, 0]]
    comb = comb[np.where(filt)]

    # compute the distances
    dx = x[comb[:, 0]] - x[comb[:, 1]]
    dy = y[comb[:, 0]] - y[comb[:, 1]]

    # filter by distance
    filt = (np.abs(dx) <= dist_limit) & (np.abs(dy) <= dist_limit)
    dx = dx[np.where(filt)]
    dy = dy[np.where(filt)]

    logger.debug(f"Determining the best dx,dy with {len(dx)} combinations.")

    return (_find_max(dx), _find_max(dy))


def match_pairs(x, y, dx, dy, tolerance=1.0):
    """Match the pairs of ordinary/extraordinary points (x, y)."""
    kd = cKDTree(list(zip(x, y)))

    px = np.array(x-dx)
    py = np.array(y-dy)

    d, ind = kd.query(list(zip(px, py)), k=1, distance_upper_bound=tolerance,
                      workers=-1)

    o = np.arange(len(x))[np.where(d <= tolerance)]
    e = np.array(ind[np.where(d <= tolerance)])
    result = Table()
    result['o'] = o
    result['e'] = e

    return result.as_array()


def quarterwave_model(psi, q, u, v, zero=0):
    """Compute polarimetry z(psi) model for quarter wavelength retarder.

    Z(psi) = Q*cos(2psi)**2 + U*sin(2psi)*cos(2psi) - V*sin(2psi)

    Parameters
    ----------
    psi: array_like
        Array of retarder positions in degrees.
    q, u, v: float
        Stokes parameters
    zero: float
        Zero position of the retarder in degrees.

    Return
    ------
    z: array_like
        Array of polarimetry values.
    """
    if zero is not None:
        psi = psi+zero  # avoid inplace modification
    psi = np.radians(psi)
    psi2 = 2*psi
    zi = q*(np.cos(psi2)**2) + u*np.sin(psi2)*np.cos(psi2) - v*np.sin(psi2)
    return zi


def halfwave_model(psi, q, u, zero=None):
    """Compute polarimetry z(psi) model for half wavelength retarder.

    Z(psi) = Q*cos(4*psi) + U*sin(4*psi)

    Parameters
    ----------
    psi: array_like
        Array of retarder positions in degrees.
    q, u: float
        Stokes parameters
    zero: float (optional)
        Zero angle of the retarder position.

    Return
    ------
    z: array_like
        Array of polarimetry values.
    """
    if zero is not None:
        psi = psi+zero  # avoid inplace modification
    psi = np.radians(psi)
    zi = q*np.cos(4*psi) + u*np.sin(4*psi)
    return zi


@dataclass
class StokesParameters:
    """Store the Stokes parameters results from dual beam polarimeters.

    Parameters
    ----------
    q, u, v: `~astropop.math.QFloat`
        Stokes parameters.
    retarder: str
        'quarterwave' or 'halfwave'
    k: float
        Normalization constant used to compute zi.
    zero: float
        Zero position of the retarder in degrees.
    zi: `~astropop.math.QFloat`
        Relative difference of fluxes between ordinary and extraordinary beams.
    psi: array_like
        Array of retarder positions in degrees.
    """

    retarder: str
    q: QFloat
    u: QFloat
    v: QFloat = None
    k: float = 1.0
    zero: float = 0.0
    zi: QFloat = None
    psi: QFloat = None

    @property
    def p(self):
        """Linear polarization level."""
        return np.hypot(self.q, self.u)

    @property
    def theta(self):
        """Angle between Stokes parameters."""
        theta = _compute_theta(self.q, self.u).nominal
        err = 28.6*self.p.std_dev/self.p.nominal
        return QFloat(theta, err, 'deg')


@dataclass
class _DualBeamPolarimetry(abc.ABC):
    """Base class for polarimetry computation."""

    retarder: str  # 'quarterwave' or 'halfwave'
    k: float = None  # global normalization constant
    zero: float = None  # zero position of the retarder. If not set, computed.
    compute_k: bool = False  # compute the normalization constant
    min_snr: float = None  # minimum signal-to-noise ratio
    psi_deviation: float = 0.1  # max deviation of retarder position
    iter_tolerance: float = 1e-5  # maximum tolerance on the iterative fitting
    max_iters: int = 100  # maximum number of iterations

    def __post_init__(self):
        if self.retarder not in ['quarterwave', 'halfwave']:
            raise ValueError(f"Retarder {self.retarder} unknown.")
        if isinstance(self.zero, (QFloat, units.Quantity)):
            self.zero = self.zero.to(units.degree).value
        if self.k is not None and self.compute_k:
            raise ValueError('k and compute_k cannot be used together.')

        # number of positions per cicle
        self._n_pos = 8 if self.retarder == 'quarterwave' else 4

    def _check_positions(self, psi):
        """Check if positions are in the correct order."""
        # Only positions multiple of 22.5 degrees are allowed
        devs = np.abs(psi/22.5 - np.round(psi/22.5, 0))*22.5
        if np.any(devs > self.psi_deviation):
            raise ValueError("Retarder positions must be multiple of 22.5 deg")

    def _calc_zi(self, f_ord, f_ext, k):
        """Compute zi from ordinary and extraordinary fluxes."""
        return (f_ord - f_ext*k)/(f_ord + f_ext*k)

    def _estimate_normalize_half(self, psi, f_ord, f_ext):
        """Estimate the normalization factor for halfwave retarder."""
        pos_in_cycle = np.mod(np.floor_divide(psi, 22.5), self._n_pos)
        pos_in_cycle = pos_in_cycle.astype(int)
        ford_mean = np.full(self._n_pos, np.nan)
        fext_mean = np.full(self._n_pos, np.nan)

        for i in range(self._n_pos):
            ford_mean[i] = np.mean(f_ord[pos_in_cycle == i])
            fext_mean[i] = np.mean(f_ext[pos_in_cycle == i])

        if np.any(np.isnan(ford_mean)) or np.any(np.isnan(fext_mean)):
            raise ValueError('Could not estimate the normalization factor.')

        return np.sum(f_ord)/np.sum(f_ext)

    def _estimate_normalize_quarter(self, psi, f_ord, f_ext, q):
        """Estimate the normalization factor for quarterwave retarder."""
        ratio = self._estimate_normalize_half(psi, f_ord, f_ext)
        q_norm = (1+0.5*q)/(1-0.5*q)
        if (ratio < 1 and q < 0) or (ratio > 1 and q > 0):
            q_norm = 1/q_norm
        return ratio*q_norm

    def _half_compute(self, psi, f_ord, f_ext):
        """Compute the Stokes params for halfwave retarder."""
        # estimate normalization factor
        if self.k is not None:
            k = self.k
        elif self.compute_k:
            k = self._estimate_normalize_half(psi, f_ord, f_ext)
        else:
            k = 1.0

        # compute Stokes params
        zi = self._calc_zi(f_ord, f_ext, k)
        q, u = self._half_fit(psi, zi)
        return StokesParameters('halfwave', q=q, u=u, v=None, k=k,
                                zero=self.zero, zi=zi)

    def _quarter_compute(self, psi, f_ord, f_ext):
        """Compute the Stokes params for quarterwave retarder."""
        # bypass normalization
        if not self.compute_k:
            k = self.k or 1.0
            zi = self._calc_zi(f_ord, f_ext, k)
            params = self._quarter_fit(psi, zi)
            return StokesParameters('quarterwave', **params, k=k,
                                    zi=zi, psi=psi)

        # iterate over k until the diference previous and current values is
        # smaller than the tolerance
        k = 1.0
        previous = {'q': 0, 'u': 0, 'v': 0}
        for i in range(self.max_iters):
            current = {'q': 0, 'u': 0, 'v': 0}
            # compute Stokes params, dict(q, u, v, zero)
            zi = self._calc_zi(f_ord, f_ext, k)
            params = self._quarter_fit(psi, zi)
            current = {k: params[k].nominal for k in previous.keys()}
            logger.debug('quarterwave iter %i: %s', i, current)
            # check if the difference is smaller than the tolerance
            if np.allclose([current[i] for i in previous.keys()],
                           [previous[i] for i in previous.keys()],
                           atol=self.iter_tolerance):
                return StokesParameters('quarterwave', **params,
                                        k=k, zi=zi, psi=psi)
            # update previous values
            previous = {i: current[i] for i in previous.keys()}

            # re-estimate k based on current q value
            k = self._estimate_normalize_quarter(psi, f_ord, f_ext,
                                                 current['q'])

        raise RuntimeError(f'Could not converge after {self.max_iters} '
                           'iterations.')

    def compute(self, psi, f_ord, f_ext, f_ord_error=None, f_ext_error=None):
        """Compute the Stokes params from ordinary and extraordinary fluxes."""
        self._check_positions(psi)
        f_ord = QFloat(f_ord, uncertainty=f_ord_error)
        f_ext = QFloat(f_ext, uncertainty=f_ext_error)

        if self.retarder == 'halfwave':
            return self._half_compute(psi, f_ord, f_ext)
        if self.retarder == 'quarterwave':
            return self._quarter_compute(psi, f_ord, f_ext)

    @abc.abstractmethod
    def _half_fit(self, psi, zi):
        """Fit the Stokes params for halfwave retarder."""

    @abc.abstractmethod
    def _quarter_fit(self, psi, zi):
        """Fit the Stokes params for quarterwave retarder."""


@dataclass
class SLSDualBeamPolarimetry(_DualBeamPolarimetry):
    """Polarimetry computation for Stokes Least Squares algorithm.

    This method is describe in [1]_ and consists in fitting the data using
    theoretical models. This method has the advantage of not need particular
    sets of retarder positions and can handle missing points. More datailed
    description of fitted equations are given in [2]_ and [3]_.

    Parameters
    ----------
    retarder: str
        Retarder type. Must be 'quarterwave' or 'halfwave'.
    k: float (optional)
        Normalization factor. If None, it is estimated from the data.
        Default is None.
    zero: float (optional)
        Zero position of the retarder in degrees. If None, it is estimated
        from the data. Defult is None. If None, it is estimated from the data
        on quarterwave retarders and will be zero for halfwave retarders.
    compute_k: bool (optional)
        Fit the normalization factor using the data. Default is False.
        Conflicts with ``k`` argument.
    min_snr: float (optional)
        Minimum signal-to-noise ratio. Points with lower SNR will be discarded.
        Default is None.
    psi_deviation: float (optional)
        Maximum deviation of the psi position from the sequence multiple
        of 22.5 degrees. Default is 0.1 degrees.

    Notes
    -----
    - The model fitting is performed by `~scipy.optimize.cure_fit` function,
      using the Trust Region Reflective ``trf`` method.
    - If ``k`` or ``zero`` arguments are passed, they won't be computed.
      Instead, the passed values will be used.
    - If ``compute_k`` is True, the value will be estimated from the data.
      For halfwave retarders, it will be estimated by the ratio of the total
      flux of the ordinary flux and the extraordinary flux in all positions
      [2]_. For quarterwave retarders, it will be estimated in a iterative
      process described in [4]_.

    References
    ----------
    .. [1] https://ui.adsabs.harvard.edu/abs/2019PASP..131b4501N
    .. [2] https://ui.adsabs.harvard.edu/abs/1984PASP...96..383M
    .. [3] https://ui.adsabs.harvard.edu/abs/1998A&A...335..979R
    .. [4] https://ui.adsabs.harvard.edu/abs/2021AJ....161..225L
    """

    def _get_fitter(self, zi):
        """Get the fitter function."""
        errors = zi.std_dev
        if np.all(errors > 0):
            fitter = partial(curve_fit, sigma=errors)
        else:
            fitter = curve_fit
            logger.info('Errors will not be considered in the fit.')
        return fitter

    def _half_fit(self, psi, zi):
        """Fit the model for halfwave retarder."""
        # change the model to do not have to fit the zero position
        model = partial(halfwave_model, zero=self.zero)
        fitter = self._get_fitter(zi)

        # fit the model
        (q, u), pcov = fitter(model, psi, zi.nominal, method='trf',
                              bounds=([-1, -1], [1, 1]))
        # Errors are the diagonal of the covariance matrix
        q_err, u_err = np.sqrt(np.diag(pcov))

        # use qfloat for fitted values
        return QFloat(q, uncertainty=q_err), QFloat(u, uncertainty=u_err)

    def _quarter_fit(self, psi, zi):
        # if zero is not set, estimate it from the data
        if self.zero is not None:
            model = partial(quarterwave_model, zero=self.zero)
            bounds = ([-1, -1, -1], [1, 1, 1])
            pnames = ['q', 'u', 'v']
        else:
            model = quarterwave_model
            bounds = ([-1, -1, -1, 0], [1, 1, 1, 180])
            pnames = ['q', 'u', 'v', 'zero']
            logger.info('Zero position not set. Computing it from the data.')

        fitter = self._get_fitter(zi)
        params, pcov = fitter(model, psi, zi.nominal, method='trf',
                              bounds=bounds)
        stokes = {pnames[i]: QFloat(params[i], uncertainty=np.sqrt(pcov[i, i]))
                  for i in range(len(pnames))}
        if len(pnames) == 3:
            stokes['zero'] = self.zero
        return stokes



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import time
import pytest
import numpy as np
from astropy.table import Table
from astropy.coordinates import SkyCoord, Angle
from astropop.catalogs.simbad import SimbadCatalog, simbad_query_id
from astropop.catalogs._online_tools import _timeout_retry, \
                                            _wrap_query_table, \
                                            astroquery_radius, \
                                            astroquery_skycoord, \
                                            get_center_radius
from astropop.testing import *
from astroquery.simbad import Simbad


def delay_rerun(*args):
    time.sleep(10)
    return True


flaky_rerun = pytest.mark.flaky(max_runs=10, min_passes=1,
                                rerun_filter=delay_rerun)
catalog_skip = pytest.mark.skipif(not os.environ.get('ASTROPOP_TEST_CATALOGS'),
                                  reason='avoid servers errors.')

sirius_coords = ["Sirius", "06h45m09s -16d42m58s", [101.28715, -16.7161158],
                 np.array([101.28715, -16.7161158]), (101.28715, -16.7161158),
                 SkyCoord(101.28715, -16.7161158, unit=('degree', 'degree'))]


@flaky_rerun
@catalog_skip
class Test_OnlineTools:
    def test_timeout_retry_error(self):
        def _only_fail(*args, **kwargs):
            assert_equal(len(args), 1)
            assert_equal(args[0], 1)
            assert_equal(len(kwargs), 1)
            assert_equal(kwargs['test'], 2)
            raise TimeoutError

        with pytest.raises(TimeoutError, match='TimeOut obtained in'):
            _timeout_retry(_only_fail, 1, test=2)

    def test_timeout_retry_pass(self):
        i = 0

        def _only_fail(*args, **kwargs):
            nonlocal i
            assert_equal(len(args), 1)
            assert_equal(args[0], 1)
            assert_equal(len(kwargs), 1)
            assert_equal(kwargs['test'], 2)
            if i < 5:
                i += 1
                raise TimeoutError
            return i

        res = _timeout_retry(_only_fail, 1, test=2)
        assert_equal(res, 5)

    def test_wrap_table(self):
        class StrObj__:
            def __init__(self, s):
                self._s = s

            def __str__(self):
                return str(self._s)

        tab = Table()
        tab['a'] = ['A3#Â'.encode('utf-8') for i in range(10)]
        tab['b'] = ['B3#Ê' for i in range(10)]
        tab['c'] = [StrObj__(i) for i in range(10)]

        _wrap_query_table(tab)

        assert_equal(len(tab), 10)
        assert_equal(tab['a'], ['A3#Â' for i in range(10)])
        assert_equal(tab['b'], ['B3#Ê' for i in range(10)])
        assert_equal(tab['c'], [f'{i}' for i in range(10)])
        assert_equal(tab['a'].dtype.char, 'U')
        assert_equal(tab['a'].dtype.char, 'U')
        assert_equal(tab['a'].dtype.char, 'U')

    def test_get_center_radius(self):
        ra = np.arange(11)
        dec = np.arange(11)
        c_ra, c_dec, rad = get_center_radius(ra, dec)
        assert_equal(c_ra, 5)
        assert_equal(c_dec, 5)
        assert_equal(rad, 10)

    @pytest.mark.parametrize('value', sirius_coords)
    def test_astroquery_skycoord_string_obj(self, value):
        skcord = astroquery_skycoord(value)
        assert_is_instance(skcord, SkyCoord)
        assert_almost_equal(skcord.ra.degree, 101.28715, decimal=3)
        assert_almost_equal(skcord.dec.degree, -16.7161158, decimal=3)

    def test_astroquery_skycoord_error(self):
        value = 'this should raise error'
        with pytest.raises(ValueError, match='could not be resolved'):
            astroquery_skycoord(value)

    def test_astroquery_radius(self):
        ang = Angle(1.0, unit='degree')
        stra = '1d'
        strb = '3600 arcsec'
        inta = 1

        assert_equal(astroquery_radius(ang), "1.0d")
        assert_equal(astroquery_radius(stra), "1.0d")
        assert_equal(astroquery_radius(strb), "1.0d")
        assert_equal(astroquery_radius(inta), "1.0d")

    def test_astroquery_radius_error(self):
        not_angle = 'not angle'
        with pytest.raises(ValueError):
            astroquery_radius(not_angle)


@flaky_rerun
@catalog_skip
class TestSimbadCatalog():
    @property
    def cat(self):
        return SimbadCatalog.copy()

    def test_simbad_catalog_get_simbad_copy(self):
        # always return a copy of the querier
        s = Simbad()

        # assign our simbad
        self.cat.simbad = s
        assert_is_not(self.cat.simbad, s)
        assert_equal(self.cat.simbad.ROW_LIMIT, 0)

    def test_simbad_catalog_set_simbad(self):
        s = Simbad()

        # ok for simbad
        self.cat.simbad = s

        # raise everything else
        for i in ['Simbad', 'simbad', 1.0, [], (), np.array([]), {}]:
            with pytest.raises(ValueError, match='is not a SimbadClass'):
                self.cat.simbad = i

    @pytest.mark.parametrize('band', ["U", "B", "V", "R", "I", "J", "H", "K",
                                      "u", "g", "r", "i", "z"])
    def test_simbad_catalog_check_filter(self, band):
        assert_true(self.cat.check_filter, band)

    def test_simbad_catalog_check_filter_errors(self):
        # default behavior is raise error
        with pytest.raises(ValueError, match='This catalog does not support'):
            self.cat.check_filter('inexisting')

        with pytest.raises(ValueError, match='This catalog does not support'):
            self.cat.check_filter('inexisting', raise_error=True)

        assert_false(self.cat.check_filter('inexisting', raise_error=False))

    def test_simbad_catalog_get_simbad(self):
        # must be free of fluxdata
        s = self.cat.get_simbad()
        assert_is_instance(s, Simbad.__class__)
        for i in s.get_votable_fields():
            assert_not_in('fluxdata', i)

        # must contain flux data
        s = self.cat.get_simbad(band='V')
        assert_is_instance(s, Simbad.__class__)
        assert_in('fluxdata(V)', s.get_votable_fields())

    def test_simbad_catalog_query_object(self):
        obj = self.cat.query_object('Sirius', band='V')
        assert_equal(obj[0]['MAIN_ID'], '* alf CMa')
        assert_equal(obj[0]['RA'], '06 45 08.9172')
        assert_equal(obj[0]['DEC'], '-16 42 58.017')
        assert_equal(obj[0]['COO_BIBCODE'], '2007A&A...474..653V')
        assert_almost_equal(obj[0]['FLUX_V'], -1.46)
        assert_equal(obj[0]['FLUX_SYSTEM_V'], 'Vega')
        assert_equal(obj[0]['FLUX_BIBCODE_V'], '2002yCat.2237....0D')

    def test_simbad_query_obj_no_band(self):
        obj = self.cat.query_object('Sirius', band=None)
        assert_equal(obj[0]['MAIN_ID'], '* alf CMa')
        assert_equal(obj[0]['RA'], '06 45 08.9172')
        assert_equal(obj[0]['DEC'], '-16 42 58.017')
        assert_equal(obj[0]['COO_BIBCODE'], '2007A&A...474..653V')
        for i in obj.colnames:
            if 'FLUX_' in i:
                raise ValueError('Simbad is getting flux data when requested '
                                 'to not.')

    @pytest.mark.parametrize('value', sirius_coords)
    def test_simbad_catalog_query_region(self, value):
        obj = self.cat.query_region(value, radius='10 arcmin',
                                    band='V')
        assert_equal(obj[0]['MAIN_ID'], '* alf CMa')
        assert_equal(obj[0]['RA'], '06 45 08.9172')
        assert_equal(obj[0]['DEC'], '-16 42 58.017')
        assert_equal(obj[0]['COO_BIBCODE'], '2007A&A...474..653V')
        assert_almost_equal(obj[0]['FLUX_V'], -1.46)
        assert_equal(obj[0]['FLUX_SYSTEM_V'], 'Vega')
        assert_equal(obj[0]['FLUX_BIBCODE_V'], '2002yCat.2237....0D')

    @pytest.mark.parametrize('value', sirius_coords)
    def test_simbad_catalog_query_region_no_band(self, value):
        obj = self.cat.query_region(value, radius='10 arcmin',
                                    band=None)
        assert_equal(obj[0]['MAIN_ID'], '* alf CMa')
        assert_equal(obj[0]['RA'], '06 45 08.9172')
        assert_equal(obj[0]['DEC'], '-16 42 58.017')
        assert_equal(obj[0]['COO_BIBCODE'], '2007A&A...474..653V')
        for i in obj.colnames:
            if 'FLUX_' in i:
                raise ValueError("Simbad is getting flux data when requested"
                                 "to not.")

    @pytest.mark.parametrize('value', sirius_coords[2:])
    def test_simbad_catalog_query_object_error(self, value):
        # Simbad query do not accept coordinates or SkyCoords
        with pytest.raises(ValueError, match="only accept object name"):
            self.cat.query_object(value)
        # but query region must pass
        self.cat.query_region(value, radius='1m')

    @pytest.mark.parametrize('radius', [Angle(0.01, unit='degree'), '0.01d',
                                        0.01, '1m', '10s', '10 arcsec'])
    def test_simbad_catalog_query_different_radius(self, radius):
        # query with different radius types
        self.cat.query_region('Sirius', radius)

    def test_simbad_catalog_filter_flux(self):
        query = Table()
        query['FLUX_V'] = np.arange(10)
        query['FLUX_ERROR_V'] = np.array([0.1]*10)
        query['FLUX_UNIT_V'] = np.array(['mag']*10)
        query['FLUX_BIBCODE_V'] = np.array(['gh2022&astropop']*10)

        flux = self.cat.filter_flux(band='V', query=query)
        assert_equal(len(flux), 4)
        assert_equal(flux[0], query['FLUX_V'])
        assert_equal(flux[1], query['FLUX_ERROR_V'])
        assert_equal(flux[2], query['FLUX_UNIT_V'])
        assert_equal(flux[3], query['FLUX_BIBCODE_V'])

    def test_simbad_catalog_filter_flux_error(self):
        query = Table()
        with pytest.raises(KeyError,
                           match='Simbad query must be performed with band'):
            self.cat.filter_flux(band='V', query=query)

    def test_simbad_catalog_filter_flux_query_none(self):
        cat = self.cat
        query = cat.query_region('Sirius', '5m', band='V')
        flux = cat.filter_flux('V')

        assert_equal(flux[0], query['FLUX_V'])
        assert_equal(flux[1], query['FLUX_ERROR_V'])
        assert_equal(flux[2], query['FLUX_UNIT_V'])
        assert_equal(flux[3], query['FLUX_BIBCODE_V'])

    def test_simbad_catalog_filter_id(self):
        # performe filter in a different class
        query = self.cat.query_region('Sirius', '5m')
        id = self.cat.filter_id(query)
        id_resolved = self.cat._id_resolve(query['MAIN_ID'])

        assert_equal(id, id_resolved)

    def test_simbad_catalog_filter_id_query_none(self):
        cat = self.cat
        query = cat.query_region('Sirius', '5m')
        idn = cat.filter_id()
        id_resolved = self.cat._id_resolve(query['MAIN_ID'])
        assert_equal(idn, id_resolved)

    def test_simbad_catalog_match_object_ids(self):
        ra = [101.28715, 88.79293875, 191.93028625]
        dec = [-16.7161158, 7.40706389, -59.68877194]
        name = ['alf CMa', 'alf Ori', 'bet Cru']
        res = self.cat.match_object_ids(ra, dec)
        assert_equal(res, name)

    @pytest.mark.parametrize('order, expect', [(None, 'alf CMa'),
                                               (['NAME'], 'Dog Star'),
                                               (['*'], 'alf CMa'),
                                               (['HIP'], 'HIP 32349'),
                                               (['HIC', 'HD'], 'HIC 32349'),
                                               (['NONE', 'HD'], 'HD 48915'),
                                               (['UBV M', 'HD'],
                                                'UBV M 12413')])
    def test_simbad_query_id(self, order, expect):
        idn = simbad_query_id(101.28715, -16.7161158, '5s', name_order=order)
        assert_equal(idn, expect)

    @pytest.mark.parametrize('coords,name', [((16.82590917, -72.4676825),
                                              'HD 6884'),
                                             ((86.46641167, -67.24053806),
                                              'LHA 120-S 61')])
    def test_simbad_query_id_non_default(self, coords, name):
        order = ['NAME', 'HD', 'HR', 'HYP', 'AAVSO', 'LHA']
        idn = simbad_query_id(*coords, '5s', name_order=order)
        assert_equal(idn, name)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Create WCS manually."""

import six
import numpy as np
from astropy.wcs import WCS

from .coords_utils import guess_coordinates


__all__ = ['wcs_from_coords']


_angles = {
    'left': 180,
    'right': 0,
    'top': 90,
    'bottom': 270
}


def wcs_from_coords(x, y, ra, dec, plate_scale, north, flip=None):
    """Create WCS based on coordinates and plate scale.

    x, y: float, pixel coordinates in image
    ra, dec: float, sky coordinates

    plate_scale in arcsec/pix

    north direction can be angles ('cw', from top axis) or
    ['right', 'top', 'left', 'bottom'] to angles [270, 0, 90, 180].

    flip means if some direction is mirrored in image.
    Can be 'ra', 'dec', 'all' or None
    The standard coordinates are ra and dec in the following order, ccw:
    W - N - E - S

    Coordinate values grow to E and N.

    Problems:
    This algorith is not good for coordinates far from crpix. But is useful
    when you cannot solve with other algorithms. (Like just one star in field).
    """
    sk = guess_coordinates(ra, dec)
    ra, dec = sk.ra.degree, sk.dec.degree

    if isinstance(north, six.string_types):
        if north in _angles.keys():
            north = _angles[north]
        else:
            raise ValueError(f'invalid value for north: {north}')

    # convert arcsec/pix to degree/pix
    plate_scale = float(plate_scale)  # ensure scalar
    plate_scale /= 3600

    # following the standard astrometry.net, all the delta informations
    # will be located inside the rotation/scale matrix pc
    deltra = plate_scale if flip == 'ra' else -plate_scale
    deltde = -plate_scale if flip == 'dec' else plate_scale
    if flip == 'all':
        north += 180

    rot = [[np.sin(np.radians(north)), -np.cos(np.radians(north))],
           [np.cos(np.radians(north)), np.sin(np.radians(north))]]

    pc = np.multiply(rot, [[deltra, deltra],
                           [deltde, deltde]])

    # and finally, we can construct the wcs
    w = WCS(naxis=2)
    w.wcs.crpix = [x, y]
    w.wcs.crval = [ra, dec]
    w.wcs.cdelt = [1, 1]
    w.wcs.ctype = ['RA---TAN', 'DEC--TAN']
    w.wcs.cunit = ['deg', 'deg']
    w.wcs.pc = pc
    # FIXME: Check why astrometry.net put pole in the crpix
    w.wcs.latpole = dec
    w.wcs.lonpole = 180

    return w



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Astrometry.net wrapper.

This module wraps the astrometry.net solve-field routine, with automatic
improvements.
"""


import os
import shutil
from subprocess import CalledProcessError
import copy
from tempfile import NamedTemporaryFile, mkdtemp

import numpy as np

from astropy.io import fits
from astropy.wcs import WCS

from .coords_utils import guess_coordinates
from ..logger import logger
from ..py_utils import check_iterable, run_command


__all__ = ['AstrometrySolver', 'solve_astrometry_xy', 'solve_astrometry_image',
           'create_xyls', 'AstrometryNetUnsolvedField']


_fit_wcs = shutil.which('fit-wcs')
_solve_field = shutil.which('solve-field')


class AstrometryNetUnsolvedField(CalledProcessError):
    """Raised if Astrometry.net could not solve the field."""

    def __init__(self, path):
        self.path = path

    def __str__(self):
        return f"{self.path}: could not solve field"


wcs_header_keys = ['CRPIX1', 'CRPIX2', 'CD1_1', 'CD1_2', 'CD2_1', 'CD2_2',
                   'CRVAL1', 'CRVAL2', 'CTYPE1', 'CTYPE2', 'CUNIT1',
                   'CUNIT2']


def clean_previous_wcs(header):
    """Clean any previous WCS keyowrds in header."""
    h = fits.Header()
    for k in header.keys():
        if k not in wcs_header_keys:
            indx = header.index(k)
            card = header.cards[indx]
            h.append(card)
    return h


class AstrometrySolver():
    """Use astrometry.net to solve the astrometry of images or list of stars.
    For convenience, all the auxiliary files will be deleted, except you
    specify to keep it with 'keep_files'.
    """
    _defaults = None

    def __init__(self, astrometry_command=_solve_field,
                 defaults=None, keep_files=False):
        # declare the defaults here to be safer
        self._defaults = {'no-plot': None, 'overwrite': None}
        if defaults is None:
            defaults = {}
        self._defaults.update(defaults)

        self._command = astrometry_command
        self._keep = keep_files
        self.logger = logger

    def _guess_coordinates(self, header, ra_key='RA', dec_key='DEC'):
        """Guess the field center based in header keys."""
        ra = header.get(ra_key)
        dec = header.get(dec_key)
        return guess_coordinates(ra, dec)

    def _guess_field_params(self, header, image_params):
        """Guess the approximate field parameters from the header.

        The estimated parameters are:
            coordinates : 'ra' and 'dec'
            plate scale : 'scale-units', 'scale-high', 'scale-low'
        """
        options = {}
        keys = image_params.keys()
        if 'ra' in keys and 'dec' in keys:
            try:
                ra = float(image_params.get('ra'))
                dec = float(image_params.get('dec'))
                self.logger.info("Usign given field coordinates: %s %s",
                                 ra, dec)
                options['ra'] = ra
                options['dec'] = dec
            except ValueError:
                self.logger.warning('Could not convert field coordinates to'
                                    ' decimal degrees. Ignoring it: %s %s',
                                    ra, dec)
        elif 'ra_key' in keys and 'dec_key' in keys:
            self.logger.info("Figuring out field center coordinates")
            try:
                coords = self._guess_coordinates(header,
                                                 image_params.get('ra_key'),
                                                 image_params.get('dec_key'))
                options['ra'] = coords.ra.degree
                options['dec'] = coords.dec.degree
            except KeyError:
                self.logger.warning("Cannot understand coordinates in"
                                    " FITS header")
        else:
            self.logger.warning("Astrometry.net will try to solve without the"
                                " field center")

        if 'pltscl' in keys:
            try:
                pltscl = image_params.get('pltscl')
                if check_iterable(pltscl):
                    pltscl = [float(i) for i in pltscl]
                else:
                    pltscl = float(image_params.get('pltscl'))
                    pltscl = [0.8*pltscl, 1.2*pltscl]
                pltscl = np.array(sorted(pltscl))
                self.logger.info("Usign given plate scale: %s", pltscl)
            except ValueError:
                self.logger.warning('Plate scale value not recognized.'
                                    ' Ignoring it. %s', pltscl)
        elif 'pltscl_key' in keys:
            self.logger.info("Figuring out the plate scale from FITS header")
            try:
                pltscl = header[image_params.get('pltscl_key')]
                pltscl = float(pltscl)
            except KeyError:
                self.logger.warning("Cannot understand plate scale in FITS"
                                    " header")
        try:
            options['scale-high'] = pltscl[1]
            options['scale-low'] = pltscl[0]
            options['scale-units'] = 'arcsecperpix'
        except NameError:
            self.logger.warning('Astrometry.net will be run without plate'
                                ' scale.')

        if 'ra' in options.keys():
            if 'radius' in keys:
                options['radius'] = float(image_params.get('radius', 1.0))
            else:
                options['radius'] = 1.0

        return options

    def solve_field(self, filename, output_file=None, wcs=False,
                    image_params=None, solve_params=None, scamp_basename=None,
                    **kwargs):
        # TODO: put some of these arguments in a config file
        """Try to solve an image using the astrometry.net.

        The image params can be:
            'ra_key' : header key for RA
            'dec_key' : header key for DEC
            'pltscl_key' : header key for plate scale
            'ra' : approximate center of the image in RA in decimal degrees
            'dec' : approximate center of the image in DEC in decimal degrees
            'pltscl' : plate scale of the image in arcsec/px
            'radius' : maximum radius to search around the center

        The solve_params are additional parameters that will be passed to
        solve-filed command. They can be:
            'no-fits2fits', 'overwrite', 'depth', 'no-plot', etc.

        'hdu' defines the number of the HDU to be solved, generally 0 (first)
        for images.

        Returns:
        --------
        header : `astropy.io.fits.Header` or `astropy.wcs.WCS`
            A header containing the solved wcs of the field and another
            information from astrometry.net. If return_wcs=True, a WCS
            object will be returned.
        """
        if image_params is None:
            image_params = {}
        if solve_params is None:
            solve_params = {}

        options = copy.copy(self._defaults)
        options.update(solve_params)

        if scamp_basename is not None:
            options['scamp-ref'] = scamp_basename + ".ref"
            options['scamp-config'] = scamp_basename + ".scamp"
            options['scamp'] = scamp_basename + ".cat"

        if output_file is not None:
            options['new-fits'] = output_file

        try:
            field_params = self._guess_field_params(fits.getheader(filename),
                                                    image_params=image_params)
        except (OSError, IOError):
            self.logger.warning('Could not guess field center and plate scale.'
                                ' Running in slow mode.')
            field_params = {}
        options.update(field_params)

        solved_header = self._run_solver(filename, params=options, **kwargs)

        if not wcs:
            return solved_header
        else:
            return WCS(solved_header, relax=True)

    def _run_solver(self, filename, params, output_dir=None, **kwargs):
        """Run the astrometry.net localy using the given params.

        STDOUT and STDERR can be stored in variables for better check after.
        """
        basename = os.path.basename(filename)
        root, _ = os.path.splitext(basename)
        if output_dir is None:
            output_dir = mkdtemp(prefix=root + '_', suffix='_astrometry.net')
            tmp_dir = True
        else:
            tmp_dir = False
        solved_file = os.path.join(output_dir, root + '.solved')

        args = [self._command, filename, '--dir', output_dir]

        for i, v in params.items():
            ndashes = 1 if len(i) == 1 else 2
            args.append(f"{ndashes * '-'}{i}")
            if v is not None:
                args.append(str(v))

        try:
            process, _, _ = run_command(args, **kwargs)
            if process.returncode != 0:
                raise CalledProcessError(process.returncode, self._command)

            # .solved file must exist and contain a binary one
            with open(solved_file, 'rb') as fd:
                if ord(fd.read()) != 1:
                    raise AstrometryNetUnsolvedField(filename)
            solved_wcs_file = os.path.join(output_dir, root + '.wcs')
            self.logger.info('Loading solved header from %s', solved_wcs_file)
            solved_header = fits.getheader(solved_wcs_file, 0)

            # remove the tree if the file is temporary and not set to keep
            if not self._keep and tmp_dir:
                shutil.rmtree(output_dir)

            return solved_header

        except CalledProcessError as e:
            if not self._keep and tmp_dir:
                shutil.rmtree(output_dir)
            raise e

        # If .solved file doesn't exist or contain one
        except (IOError, AstrometryNetUnsolvedField):
            if not self._keep and tmp_dir:
                shutil.rmtree(output_dir)
            raise AstrometryNetUnsolvedField(filename)


def create_xyls(fname, x, y, flux, imagew, imageh, header=None, dtype='f8'):
    '''
    Create and save the xyls file to run in astrometry.net

    Parameters:
    -----------
    fname : str
        The path to save the .xyls file.
    x : array_like
        X coordinates of the sources.
    y : array_like
        Y coordinates of the sources.
    flux : array_like
        Estimated flux of the sources.
    imagew : float or int
        Width of the original image. `IMAGEW` header field of .xyls
    imageh : float or int
        Height of the original image. `IMAGEH` header field of .xyls
    dtype : `numpy.dtype`, optional
        Data type of the fields. Default: 'f8'
    '''
    head = {'IMAGEW': imagew,
            'IMAGEH': imageh}
    sort = np.argsort(flux)
    xyls = np.array(list(zip(x, y, flux)),
                    np.dtype([('x', dtype), ('y', dtype), ('flux', dtype)]))
    f = fits.HDUList([fits.PrimaryHDU(header=header),
                      fits.BinTableHDU(xyls[sort[::-1]],
                                       header=fits.Header(head))])
    logger.debug('Saving .xyls to %s', fname)
    f.writeto(fname)


def solve_astrometry_xy(x, y, flux, image_header, image_width, image_height,
                        return_wcs=False, image_params=None,
                        **kwargs):
    """Solve astrometry from a (x,y) sources list.

    Notes
    -----
    - image_params are:
      * pltscl: plate scale (arcsec/px)
      * ra: right ascension (decimal degrees)
      * dec: declination (decimal degrees)
      * pltscl_key: header key for plate scale
      * ra_key: header key for right ascension
      * dec_key: header key for declination
      * radius: maximum search radius
    """
    if image_params is None:
        image_params = {}
    image_header = clean_previous_wcs(image_header)
    f = NamedTemporaryFile(suffix='.xyls')
    create_xyls(f.name, x, y, flux, image_width, image_height,
                header=image_header)
    solver = AstrometrySolver()
    return solver.solve_field(f.name, wcs=return_wcs,
                              image_params=image_params, **kwargs)


def solve_astrometry_image(filename, return_wcs=False, image_params=None,
                           **kwargs):
    """
    image_params are:
        pltscl: plate scale (arcsec/px)
        ra: right ascension (decimal degrees)
        dec: declination (decimal degrees)
        pltscl_key: header key for plate scale
        ra_key: header key for right ascension
        dec_key: header key for declination
        radius: maximum search radius
    """
    if image_params is None:
        image_params = {}
    solver = AstrometrySolver()
    return solver.solve_field(filename, wcs=return_wcs,
                              image_params=image_params, **kwargs)


def solve_astrometry_hdu(hdu, return_wcs=False, image_params=None,
                         **kwargs):
    """
    image_params are:
        pltscl: plate scale (arcsec/px)
        ra: right ascension (decimal degrees)
        dec: declination (decimal degrees)
        pltscl_key: header key for plate scale
        ra_key: header key for right ascension
        dec_key: header key for declination
        radius: maximum search radius
    """
    if image_params is None:
        image_params = {}
    hdu.header = clean_previous_wcs(hdu.header)
    f = NamedTemporaryFile(suffix='.fits')
    hdu = fits.PrimaryHDU(hdu.data, header=hdu.header)
    hdu.writeto(f.name)
    solver = AstrometrySolver()
    return solver.solve_field(f.name, wcs=return_wcs,
                              image_params=image_params, **kwargs)


def fit_wcs(x, y, ra, dec, image_width, image_height, sip=False,
            command=_fit_wcs, **kwargs):
    """Run astrometry.net fit-wcs command

    sip is sip order, int
    """
    solved_wcs_file = NamedTemporaryFile(prefix='fitwcs', suffix='.wcs')
    tmp_table = NamedTemporaryFile(prefix='fitwcs_xyrd', suffix='.fits')

    xyrd = np.array(list(zip(x, y, ra, dec)),
                    np.dtype([('FIELD_X', 'f8'), ('FIELD_Y', 'f8'),
                              ('INDEX_RA', 'f8'), ('INDEX_DEC', 'f8')]))
    f = fits.HDUList([fits.PrimaryHDU(),
                      fits.BinTableHDU(xyrd)])
    logger.debug('Saving xyrd to %s', tmp_table.name)
    f.writeto(tmp_table.name)

    args = [command]
    args += ['-c', tmp_table.name]

    if sip:
        args += ['-s', str(sip), '-W', str(image_width), '-H',
                 str(image_height)]
    args += ['-o', solved_wcs_file.name]

    try:
        process, _, _ = run_command(args, **kwargs)
        if process.returncode != 0:
            raise CalledProcessError(process.returncode, args)
        logger.info('Loading solved header from %s', solved_wcs_file.name)
        solved_header = fits.getheader(solved_wcs_file.name, 0)
    except Exception as e:
        raise AstrometryNetUnsolvedField("Could not fit wcs to this lists."
                                         f" Error: {e}")

    return WCS(solved_header, relax=True)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
import shlex
from astropop.py_utils import string_fix, process_list, \
                              check_iterable, batch_key_replace, \
                              run_command, IndexedDict, check_number, \
                              broadcast
import numpy as np

from astropop.testing import *
from astropop.logger import logger, log_to_list


@pytest.mark.parametrize('v, exp', [(1, True),
                                    (1.5, True),
                                    ('2', False),
                                    ('2.5', False),
                                    (1+3j, False),
                                    (5j, False),
                                    ('A', False),
                                    ('AB', False),
                                    ([1, 2, 3], False),
                                    (np.array([1, 2]), False),
                                    (np.int('3'), True),
                                    (np.float(3), True),
                                    (False, False),
                                    ((1, 2, 3), False)])
def test_check_number(v, exp):
    if exp:
        assert_true(check_number(v))
    else:
        assert_false(check_number(v))


@pytest.mark.parametrize("inp, enc, res", [("a!^1ö~[😀", None, "a!^1ö~[😀"),
                                           ("a!^1ö~[😀", "utf-8", "a!^1ö~[😀"),
                                           ("a!1[", 'latin-1', "a!1["),
                                           (b'bytes', None, 'bytes'),
                                           (42, None, "42")])
def test_string_fix(inp, enc, res):
    if enc is not None:
        inp = inp.encode(enc)
        assert_equal(string_fix(inp, enc), res)
    else:
        assert_equal(string_fix(inp), res)


class Test_RunCommand():
    com = (["bash", "-c", 'for i in {1..10}; do echo "$i"; sleep 0.1; done'],
           "bash -c 'for i in {1..10}; do echo \"$i\"; sleep 0.1; done'")
    com2 = ("bash -c 'echo \"this is an error\" 1>&2'",
            ["bash", "-c", 'echo "this is an error" 1>&2'])

    @pytest.mark.parametrize('com', com)
    def test_run_command(self, com):
        stdout = []
        stderr = []
        _, out, err = run_command(com, stdout=stdout, stderr=stderr,
                                  stdout_loglevel='WARN')
        assert_is(out, stdout)
        assert_is(err, stderr)
        assert_equal(stdout, [str(i) for i in range(1, 11)])
        assert_equal(stderr, [])

    @pytest.mark.parametrize('com', com2)
    def test_run_command_stderr(self, com):
        stdout = []
        stderr = []
        _, out, err = run_command(com, stdout=stdout, stderr=stderr,
                                  stdout_loglevel='WARN')
        assert_is(out, stdout)
        assert_is(err, stderr)
        assert_equal(stdout, [])
        assert_equal(stderr, ['this is an error'])

    @pytest.mark.parametrize('com', com)
    def test_logging(self, com):
        logger.setLevel('DEBUG')
        logl = []
        expect_log = []
        if not isinstance(com, list):
            com = shlex.split(com)
            # expect_log += ['Converting string using shlex']
        logcmd = com
        logcmd = " ".join(logcmd)
        expect_log += [f"Runing: {logcmd}"]
        expect_log += list(range(1, 11))
        expect_log += [f"Done with process: {logcmd}"]

        lh = log_to_list(logger, logl)
        stdout = []
        stderr = []
        _, out, err = run_command(com, stdout=stdout, stderr=stderr,
                                  stdout_loglevel='WARN')
        assert_is(out, stdout)
        assert_is(err, stderr)
        assert_equal(stdout, [str(i) for i in range(1, 11)])
        assert_equal(stderr, [])
        assert_equal(logl, expect_log)
        logger.removeHandler(lh)

    @pytest.mark.parametrize('com', com2)
    def test_logging_err(self, com):
        logger.setLevel('DEBUG')
        logl = []
        expect_log = []
        if not isinstance(com, list):
            com = shlex.split(com)
            # expect_log += ['Converting string using shlex']
        logcmd = com
        logcmd = " ".join(logcmd)
        expect_log += [f"Runing: {logcmd}"]
        expect_log += ['this is an error']
        expect_log += [f"Done with process: {logcmd}"]

        lh = log_to_list(logger, logl)
        stdout = []
        stderr = []
        _, out, err = run_command(com, stdout=stdout, stderr=stderr,
                                  stdout_loglevel='DEBUG',
                                  stderr_loglevel='ERROR')
        assert_is(out, stdout)
        assert_is(err, stderr)
        assert_equal(stdout, [])
        assert_equal(stderr, ['this is an error'])
        assert_equal(logl, expect_log)
        logger.removeHandler(lh)


class Test_ProcessList():
    def test_process_list(self):
        def dummy_func(i):
            i = 1
            return i
        a = np.zeros(20)
        b = np.ones(20)
        c = process_list(dummy_func, a)
        assert_equal(b, c)
        assert_false(np.array_equal(a, c))

    def test_process_list_with_args(self):
        def dummy_func(i, a, b):
            return (i+a)*b
        i_array = np.arange(20)
        a_val = 2
        b_val = 3
        res = process_list(dummy_func, i_array, a_val, b=b_val)
        assert_equal((i_array+a_val)*b_val, res)


class Test_CheckIterable():
    def test_check_iterabel_array(self):
        a = [1, 2, 3, 4, 5]
        assert_true(check_iterable(a))

    def test_check_iterabel_string(self):
        a = '12345'
        assert_false(check_iterable(a))

    def test_check_iterabel_nparray(self):
        a = np.zeros(20)
        assert_true(check_iterable(a))

    def test_check_iterabel_number(self):
        a = 10
        assert_false(check_iterable(a))

    def test_check_iterabel_range(self):
        a = range(10)
        assert_true(check_iterable(a))

    def test_check_iterabel_dict(self):
        a = dict(a=1, b=2, c=3, d=4)
        assert_true(check_iterable(a))
        assert_true(check_iterable(a.items()))
        assert_true(check_iterable(a.keys()))
        assert_true(check_iterable(a.values()))


class Test_BatchKeyReplace():
    def test_batch_key_replace(self):
        dic1 = {'a': '{b} value', 'b': '6{c}', 'c': 2}
        batch_key_replace(dic1)
        assert_equal(dic1['a'], '62 value')

    def test_batch_key_replace_list(self):
        dic1 = {'a': '{b} value', 'b': ['6{c}', '4{d}'], 'c': 1, 'd': 2}
        batch_key_replace(dic1)
        assert_equal(dic1['a'], "['61', '42'] value")


class Test_Broadcast():
    def test_broadcast(self):
        a = np.arange(10)
        b = np.arange(1, 11)
        c = 2

        bc = broadcast(a, b, c)
        iterator = iter(bc)

        indx = 0
        for i in range(10):
            assert_equal(next(iterator), [a[indx], b[indx], c])
            indx += 1

        with pytest.raises(StopIteration):
            next(iterator)

    def test_broadcast_empty(self):
        with pytest.raises(ValueError):
            broadcast()

    def test_broadcast_wrong_shape(self):
        a = np.arange(10)
        b = np.arange(5)

        with pytest.raises(ValueError):
            broadcast(a, b)

    def test_broadcast_only_scalars(self):
        a = 1
        b = 2
        c = 3

        bc = broadcast(a, b, c)

        for i in bc:
            assert_equal(i, [a, b, c])

    def test_broadcast_superpass_32_limit(self):
        arr = [np.arange(10)]*64
        bc = broadcast(*arr)
        assert_equal(len(bc), 10)
        it = iter(bc)

        for i in range(10):
            assert_equal(next(it), [i]*64)

    def test_broadcast_iters_only_scalars(self):
        bc = broadcast(1, 2, 3)
        assert_equal(bc.iters, [[1], [2], [3]])

    def test_broadcast_iters(self):
        bc = broadcast(np.arange(10), 3, 2)

        assert_equal(bc.iters, [np.arange(10), [3]*10, [2]*10])

class Test_IndexedDict():
    def test_indexeddict_create(self):
        d = dict(a=1, b=2, c=3)
        i = IndexedDict(a=1, b=2, c=3)
        assert_is_instance(i, dict)
        assert_equal(len(d), len(i))
        # Python 3.6 and above ensure items order
        assert_equal(list(d.keys()), list(i.keys()))
        assert_equal(list(d.values()), list(i.values()))
        assert_equal(i, d)

    def test_indexeddict_insert_at(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        a.insert_at(2, 'e', 5)
        assert_equal(a, {'a': 1, 'b': 2, 'e': 5, 'c': 3, 'd': 4})

    def test_indexeddict_insert_at_first(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        a.insert_at(0, 'e', 5)
        assert_equal(a, {'e': 5, 'a': 1, 'b': 2, 'c': 3, 'd': 4})

    def test_indexeddict_insert_at_last(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        a.insert_at(4, 'e', 5)
        assert_equal(a, {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5})

    def test_indexeddict_insert_at_away(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        a.insert_at(42, 'e', 5)
        assert_equal(a, {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5})

    def test_indexeddict_insert_at_negative(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        a.insert_at(-2, 'e', 5)
        assert_equal(a, {'a': 1, 'b': 2, 'c': 3, 'e': 5, 'd': 4})

    def test_indexeddict_after(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        a.insert_after('b', 'e', 5)
        assert_equal(a, {'a': 1, 'b': 2, 'e': 5, 'c': 3, 'd': 4})

    def test_indexeddict_before(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        a.insert_before('b', 'e', 5)
        assert_equal(a, {'a': 1, 'e': 5, 'b': 2, 'c': 3, 'd': 4})

    def test_indexeddict_existing_before_before(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        a.insert_before('b', 'c', 3)
        assert_equal(a, {'a': 1, 'c': 3, 'b': 2, 'd': 4})

    def test_indexeddict_existing_after_before(self):
        a = IndexedDict(a=1, b=2, c=3, d=4, e=5)
        a.insert_before('e', 'c', 4)
        assert_equal(a, {'a': 1, 'b': 2, 'd': 4, 'c': 4, 'e': 5})

    def test_indexeddict_existing_before_after(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        a.insert_after('b', 'c', 3)
        assert_equal(a, {'a': 1, 'c': 3, 'b': 2, 'd': 4})

    def test_indexeddict_existing_after_after(self):
        a = IndexedDict(a=1, b=2, c=3, d=4, e=5)
        a.insert_after('e', 'c', 4)
        assert_equal(a, {'a': 1, 'b': 2, 'd': 4, 'c': 4, 'e': 5})

    def test_indexeddict_first(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        a.insert_before('a', 'e', 5)
        assert_equal(a, {'e': 5, 'a': 1, 'b': 2, 'c': 3, 'd': 4})

    def test_indexeddict_last(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        a.insert_after('d', 'e', 5)
        assert_equal(a, {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5})

    @pytest.mark.parametrize('val, res', [('a', 0), ('b', 1),
                                          ('c', 2), ('d', 3)])
    def test_indexeddict_index(self, val, res):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        assert_equal(a.index(val), res)

    def test_indexeddict_invalid_key(self):
        a = IndexedDict(a=1, b=2, c=3, d=4)
        with pytest.raises(KeyError):
            a.index('e')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Put a `unit` property in classes."""

from astropy import units


def unit_property(cls):
    """Add a `unit` property to a class."""
    def _unit_getter(obj):
        if obj._unit is None:
            return units.dimensionless_unscaled
        return obj._unit

    def _unit_setter(obj, value):
        if value is None or units.Unit(value) == units.dimensionless_unscaled:
            obj._unit = None
        else:
            obj._unit = units.Unit(value)

    cls._unit = None
    cls.unit = property(_unit_getter, _unit_setter,
                        doc="Physical unit of the data.")
    return cls



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utilities for loading data as FrameData."""

import os
import numpy as np
from astropy.io import fits
from astropy.nddata import CCDData
from astropy import units as u

from .framedata import FrameData
from .compat import _extract_ccddata, _extract_fits, imhdus
from .memmap import MemMapArray


__all__ = ['check_framedata', 'read_framedata']


_fits_kwargs = ['hdu', 'unit', 'hdu_uncertainty',
                'hdu_mask', 'unit_key']


def read_framedata(obj, copy=False, **kwargs):
    """Read an object to a FrameData container.

    Parameters
    ----------
    obj: any compatible, see notes
      Object that will be readed to the FrameData.
    copy: bool (optional)
      If the object is already a FrameData, return a copy instead of the
      original one.
      Default: False

    Returns
    -------
    frame: `FrameData`
      The readed FrameData object.

    Notes
    -----
    - If obj is a string or `~pathlib.Path`, it will be interpreted as a file.
      File types will be checked. Just FITS format supported now.
    - If obj is `~astropy.io.fits.HDUList`, `~astropy.io.fits.HDUList` or
      `~astropy.nddata.CCDData`, they will be properly translated to
      `FrameData`.
    - If numbers or `~astropop.math.physical.QFloat`,
      `~astropy.units.Quantity`, they will be translated to a `FrameData`
      without metadata.
    """
    if isinstance(obj, FrameData):
        if copy:
            obj = obj.copy(**kwargs)
    elif isinstance(obj, CCDData):
        obj = FrameData(**_extract_ccddata(obj), **kwargs)
    elif isinstance(obj, (str, bytes, os.PathLike, fits.HDUList, *imhdus)):
        # separate kwargs to be sent to extractors
        fits_kwargs = {}
        for k in _fits_kwargs:
            if k in kwargs.keys():
                fits_kwargs[k] = kwargs.pop(k)
        obj = FrameData(**_extract_fits(obj, **fits_kwargs), **kwargs)
    elif isinstance(obj, (np.ndarray, MemMapArray)):
        if isinstance(obj, u.Quantity):
            obj = FrameData(obj.value, unit=obj.unit, **kwargs)
        else:
            obj = FrameData(obj, **kwargs)
    elif obj.__class__.__name__ == "QFloat":
        # if not do this, a cyclic dependency breaks the code.
        obj = FrameData(obj.nominal, unit=obj.unit,
                        uncertainty=obj.uncertainty, **kwargs)
    else:
        raise TypeError(f'Object {obj} is not compatible with FrameData.')

    return obj


check_framedata = read_framedata
